const seedrandom = require('seedrandom');

const Matter = require("matter-js");

const decomp = require('poly-decomp');
Matter.Common.setDecomp(decomp);

const utils = require("./utils.js");
const arrangements = require("./arrangements.js");
const brains = require("./brains/cogitator.js");
const loops = require("./loops.js");

const robotman = require('../robotman.js');
const simlogger = require("./simlogger.js");

// note: matterjs "angle" uses clockwise rotation from +ve x axis
// based on: "The rotation angle, clockwise in radians" for canvas
// https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/rotate

const angle360 = Math.PI * 2;
const angle180 = Math.PI;
const angle90 = Math.PI / 2;

const angleNorth = Math.PI * 3 / 2;
const angleSouth = Math.PI / 2;
const angleWest = Math.PI;
const angleEast = 0;

const robot_polygon_path = "0 100 20 50 50 50 " +
    "100 0 150 0 160 100 150 200 100 200 50 150 20 150";

let IIDCounter = -1;

let DEBUG_INJECT_INTO_ROBOT = false;

const set_debug_inject_into_robot = (flag) => {
    DEBUG_INJECT_INTO_ROBOT = Boolean(parseInt(flag) || 0);
    console.log(
        '[sim:GLOBAL] DEBUG_INJECT_INTO_ROBOT =',
        DEBUG_INJECT_INTO_ROBOT);
}

// dev: use a factory pattern for parallel runs
function Factory(options) {

    // dev: scope update loops for mass halt
    let Loops = loops.Factory();

    IIDCounter++;

    const settings = {
        sim: {
            IID: IIDCounter,
            created: new Date(),
            maxRounds: -1,
            maxRuntime: 86400,
            seed: "fixedseed", // usually changed via options
            realtime: true,
            physicsFPS: 60, // server physics framerate
            updateFPS: 18, // client framerate
            autoRestart: 5.0,
            numOfRobots: 4,
            robotScale: 0.15,
            defaultBrain: brains.driver.debug_serverside,
            projectileSize: 3,
            projectileLeadOffset: -15,
            locked: false, // prevent reinitialisation
            // derived from other settings, or set later
            physicsRate: false, // re: physicsFPS
            updateRate: false, // re: updateFPS
            PRNG: false, // prng seeded with seed
            identicalRounds: false, // true = re-seed PRNG on each round
            ignoreSingleTeamEndsRound: true, // prevents multi robot, single team rounds from ending
        },
        arena: {
            width: 900,
            height: 900,
            wall: 10,
        },
        rules: {
            arrangement: {
                type: "raytrace",
                options: {
                    Sides: [0, 1, 2, 3],
                    angleOffset: Math.PI,
                    snapAngle: angle90
                }
            },
            limit: {
                round: 60,
                lastrobot: 3.0,
                robotMaxHealth: 200,
                scanLenMin: 50,
                scanLenMax: false, // derived
                scanScaleMax: 10,
                scanScaleMin: 1
            },
            damage: {
                wall: 1,
                robot: 2,
                projectile: 5,
                executorError: 5,
            },
            physics: {
                robotFrictionAir: 0.03,
                robotMaxForce: 0.001,
                robotMaxRotate: 0.272,
                projectilePerFrame: 0.5,
                projectileForce: 0.001,
                projectileRecoilForce: 0.001,
                projectileMinSpeed: 10
            },
            initial: {
                robotScanDefault: 150
            }
        },
        brains: []
    };

    const assign_keys = (keys, target, source) => {
        for (let key of keys) {
            let merge = source[key] || {};
            if (typeof(merge) !== "object") throw "options." + key;
            target[key] = Object.assign(target[key], merge);
        }
    }

    options = options || {};

    assign_keys(
        ["sim", "arena", "brains"],
        settings,
        options);

    assign_keys(
        Object.keys(settings.rules),
        settings.rules,
        options.rules || {});

    // derived settings

    settings.sim.physicsRate = (
        1000 / settings.sim.physicsFPS); // == ms/frame

    settings.sim.updateRate = (
        1000 / settings.sim.updateFPS);

    settings.rules.limit.scanLenMax = Math.max(
        settings.arena.width, settings.arena.height) / 2;

    settings.rules.physics.projectileRate = (
        settings.sim.physicsFPS * settings.rules.physics.projectilePerFrame);

    // initialise pseudo-random number generator
    // used for deterministic battle states

    const seed_prng = () => {
        if (settings.sim.identicalRounds || !settings.sim.PRNG) {
            settings.sim.PRNG = seedrandom(settings.sim.seed);
            console.log(
                "[sim:" + settings.sim.IID + "::seed_prng]",
                "length:", settings.sim.seed.length);
        }
    }

    // debug states

    const debug = {
        state           : 0,
        randomForward   : 0.5,
        randomHeading   : 0.005,
        scanFire        : true,
        PRNG            : () => {return settings.sim.PRNG();}
    };

    switch(debug.state) {
        case 0:
            // production
            break
        case 1:
            // debug: slow projectiles
            settings.sim.numOfRobots = 2;
            settings.rules.physics.projectileForce = 0.00002;
            settings.rules.physics.projectileMinSpeed = 0;
            debug.randomHeading = 0;
            debug.randomForward = 0;
            break;
        case 2:
            // debug: single dumb robot
            settings.sim.numOfRobots = 1;
            settings.rules.limit.lastrobot = 1000;
            debug.randomHeading = 0;
            debug.randomForward = 0;
            break;
        case 3:
            // debug: 3 robots - shooty
            settings.sim.numOfRobots = 3;
            debug.randomHeading = 0;
            debug.randomForward = 0;
            break;
        case 4:
            // debug: 2 robots - dumb
            settings.sim.numOfRobots = 2;
            debug.randomHeading = 0;
            debug.randomForward = 0;
            debug.scanFire = false;
            break;
        case 5:
            // debug: 24 fragile bots
            settings.sim.numOfRobots = 24;
            settings.rules.limit.robotMaxHealth = 10;
            break;
        case 6:
            // debug: 16 dumb bots
            settings.sim.numOfRobots = 16;
            debug.randomHeading = 0;
            debug.randomForward = 0;
            break;
        case 7:
            // debug: increased chances of quick tie
            settings.sim.numOfRobots = 24;
            settings.rules.damage.wall = 5;
            settings.rules.damage.robot = 0;
            settings.rules.damage.projectile = 200;
            break;
        default:
        // nothing
    }

    // start of logic

    const collision_targets = 0x0002;

    // dev: state is sent to clients - no private/secret values!
    let state = {
        started: false,
        realtime: settings.sim.realtime,
        arena: settings.arena,
        round: {
            end: false,
            remaining: false,
            results: false
        },
        count: {
            rounds: 0,
            frames: 0,
            robots: 0,
            teams: 0,
            seconds: 0,
        },
        rankings: false,
        stopped: false,
    };

    let Engine = false,
        // dev: Entities is sent to clients
        Entities = {
            sensors: [],
            robots: [],
            projectiles: [],
            walls: []
        },
        BodyById = {},
        ScanByRobot = {},
        socketio = false,
        socketroom = false,
        simlog_id = "unset-" + (new Date()).getTime();

    let Timeouts = {
        "recovery": false,
        "autorestart": false
    };

    /* socketio */

    const set_socket_io = socks => {
        // simulation needs socketio to broadcast state to clients
        socketio = socks;
        console.log(
            "[sim:" + settings.sim.IID + "::set_socket_io]",
            "linked");
    };

    const set_socket_room = roomid => {
        socketroom = roomid;
        console.log(
            "[sim:" + settings.sim.IID + "::set_socket_room]",
            roomid);
    };

    const broadcast = (message, metadata) => {
        if (!socketio) return false;
        socketio.to(socketroom).emit(message, metadata);
    };

    /* logger */

    const set_simlog_id = str => {
        simlog_id = str;
    }

    /* vertex and state serialisers */

    const toGenericVertexes = e => ({
        id              : e.id,
        edges           : e.vertices.map(({x, y}) => ({x, y})),
        center          : [e.position.x, e.position.y],
        angle           : utils.clamped_angle(e.angle),
        fillStyle       : e.render.fillStyle,
        maxDimension    : e.maxDimension || false
    });

    const toScanRanges = e => ({
        id              : e.id,
        edges           : e.vertices.map(({x, y}) => ({x, y})),
        center          : [e.position.x, e.position.y],
        angle           : utils.clamped_angle(e.angle),
        fillStyle       : e.render.fillStyle,
        scanDistance    : e.scanDistance
    });

    const toRobots = e => ({
        id              : e.id,
        edges           : e.vertices.map(({x, y}) => ({x, y})),
        center          : [e.position.x, e.position.y],
        angle           : utils.clamped_angle(e.angle),
        active          : e.active,
        fillStyle       : e.render.fillStyle,
        entityName      : e.entityName,
        entityTeam      : e.entityTeam,
        entityWidth     : e.entityWidth,
        entityHeight    : e.entityHeight,
        maxDimension    : e.maxDimension || false,
        brainStatus     : e.brainStatus,
        health:
            100 - Math.floor(e.damage / e.maxHealth * 100),
        /*
        text: {
            "robotHeading":
                utils.abs_bearing(e.angle, 1) +
                    " (" + utils.clamped_angle(e.angle, 1) + ")" +
                " / " + (e.desiredAngle === false ?
                    "X" : utils.abs_bearing(e.desiredAngle, 1)),
            "scannerSummary":
                "A:" + (e.scanner.ALL || []).length + " " +
                "R:" + (e.scanner.robot || []).length + " " +
                "P:" + (e.scanner.projectile || []).length + " " +
                "W:" + (e.scanner.wall || []).length,
            "scannerClosestRobot":
                e.scanner.robot.length > 0 ? e.scanner.robot[0].id : "none"
        },
        */
        debugScannerALL : e.scanner.ALL
    });

    const toCircles = e => ({
        id      : e.id,
        radius  : e.circleRadius,
        center  : [e.position.x, e.position.y]
    });

    /*  helpers */

    const radians_to_center = (body) => {
        // angle to arena center from given body

        return utils.angle_to(
            body.position.x, body.position.y,
            (settings.arena.width / 2), (settings.arena.height / 2));
    };

    const damage_rgb = robot => {
        // return RGB dict based on [0,255] / robot.damage

        let ratio = 1.0;

        if (typeof robot == "object") {
            ratio = robot.damage / robot.maxHealth
        } else {
            ratio = parseFloat(robot);
        }

        return utils.color_gradient(ratio,
            {"red": 0, "green": 255, "blue": 0},
            {"red": 255, "green": 255, "blue": 0},
            {"red": 255, "green": 0, "blue": 0});
    };

    const only_pair = (pair, entityType1, entityType2) => {
        // filter matterjs collision pairs and guarantee specific order

        var parentA = pair.bodyA.parent,
            parentB = pair.bodyB.parent;

        if (parentA.entityType == entityType1
                && parentB.entityType == entityType2)
            return [parentA, parentB];

        if (parentA.entityType == entityType2
                && parentB.entityType == entityType1)
            return [parentB, parentA];

        return false;
    }

    /*  initialisation and setup */

    const arena_wall = (x, y, w, h, options) => {
        return Matter.Bodies.rectangle(
            x, y, w, h,
            Object.assign({
                // physics properties (matterjs)
                isStatic: true,
                render: {
                    fillStyle: "gray"
                },
                // custom properties
                "entityType": "wall"
            }, options || {})
        )
    };

    const setup_robots = () => {
        let vertices = Matter.Vertices.fromPath(robot_polygon_path);

        Matter.Vertices.scale(
            vertices,
            settings.sim.robotScale,
            settings.sim.robotScale);

        let ys = vertices.map(v => v.y),
            xs = vertices.map(v => v.x),
            robotWidth = Math.max(...xs) - Math.min(...xs),
            robotHeight = Math.max(...ys) - Math.min(...ys),
            robotMaxDimension = Math.max(robotWidth, robotHeight);

        // dev: useful for sprite sizing
        // console.log("[setup] robot vertex:", robotWidth, robotHeight);

        // determine start positions

        var robotStartPositions = [],
            minGap = 50,
            diagonal = Math.sqrt(
                Math.pow(robotWidth, 2) + Math.pow(robotHeight, 2)),
            spacer = Math.max(diagonal, minGap),
            arena_cx = (settings.arena.width / 2),
            arena_cy = (settings.arena.height / 2),
            allowed_x = settings.arena.width - (spacer * 2),
            allowed_y = settings.arena.height - (spacer * 2);

        if (settings.sim.numOfRobots == 1) {
            robotStartPositions.push([
                settings.arena.width / 2,
                settings.arena.height / 2,
                0 // "natural" angle - pointing to right
            ]);
        } else {
            robotStartPositions = (
                arrangements[settings.rules.arrangement.type](
                    settings.sim.numOfRobots, arena_cx, arena_cy,
                    allowed_x, allowed_y,
                    settings.rules.arrangement.options || {}));
        }

        // generate the robot physic bodies

        let unique_names = {},  // tracks _entityName
            unique_ids = {};    // tracks _entityID

        let robots = [...Array(settings.sim.numOfRobots)].map((v, idx) => {
            let robotPosition = robotStartPositions[idx],
                robotBrain = settings.brains[idx],
                failsafes = [];

            if (!robotPosition) {
                robotPosition = [
                    settings.sim.PRNG() * allowed_x + spacer,
                    settings.sim.PRNG() * allowed_y + spacer,
                    settings.sim.PRNG() * angle360
                ];
                failsafes.push("randompos");
            }

            let _entityID = 'R' + (idx + 1),
                _entityName = "Debug " + (idx + 1),
                _entityTeam = "Team " + (idx % 8 + 1);

            let _acquireSource = false;

            if (robotBrain) {
                // supplied externally
                if (robotBrain.version == 1) {
                    if ('id' in robotBrain) {
                        // dev: different than robot.id
                        _entityID = robotBrain.id;
                        _entityName = robotBrain.id;
                    }

                    if ('name' in robotBrain) _entityName = robotBrain.name;
                    if ('team' in robotBrain) _entityTeam = robotBrain.team;

                    _acquireSource = robotBrain.acquire_source || false;

                    robotBrain = robotBrain.brain;
                }
            } else {
                // use default brain if not specified
                robotBrain = settings.sim.defaultBrain;
                failsafes.push("defaultBrain");
            }

            // track unique robot names
            // appends number on duplicates
            // dev: names can duplicate when using identical server-side brains
            if (_entityName in unique_names) {
                unique_names[_entityName]++;
                _entityName += ' ' + unique_names[_entityName];
                console.log(
                    "[sim:" + settings.sim.IID + "::setup_robots]",
                    idx + ", duplicate robot name replaced", _entityName);
            } else {
                unique_names[_entityName] = 1;
            }

            // track unique robot ids
            // suffix with [number] on duplicates
            // dev: square brackets are reserved for this
            // dev: ids can duplicate when using identical server-side brains
            if (_entityID in unique_ids) {
                unique_ids[_entityID]++;
                _entityID += '[' + unique_ids[_entityID] + ']';
                console.log(
                    "[sim:" + settings.sim.IID + "::setup_robots]",
                    idx + ", duplicate robot id replaced", _entityID);
            } else {
                unique_ids[_entityID] = 1;
            }

            if (failsafes.length) {
                console.log(
                    "[sim:" + settings.sim.IID + "::setup_robots]",
                    idx + ", loaded failsafes:", failsafes);
            }

            return Matter.Bodies.fromVertices(
                robotPosition[0], robotPosition[1], vertices,
                {   // physics properties (matterjs)
                    isStatic        : false,
                    restitution     : 0.95,
                    frictionAir     : settings.rules.physics.robotFrictionAir,
                    render: {
                        fillStyle   : damage_rgb(0)
                    },
                    collisionFilter: {
                        category    : collision_targets
                    },
                    // use for external reference, optional
                    entityID        : _entityID,
                    entityGetSrc    : _acquireSource,
                    // custom identification - visible to robots
                    entityName      : _entityName,
                    entityTeam      : _entityTeam,
                    entityType      : "robot",
                    // custom settings
                    entityWidth     : robotWidth,
                    entityHeight    : robotHeight,
                    maxDimension    : robotMaxDimension,
                    maxHealth       : settings.rules.limit.robotMaxHealth,
                    maxForce        : settings.rules.physics.robotMaxForce,
                    maxRotate       : settings.rules.physics.robotMaxRotate,
                    // custom states
                    active          : true,
                    memory          : {},
                    brain           : robotBrain,
                    brainStatus     : true,
                    damage          : 0,
                    tkoAt           : -1,
                    desiredScan     : settings.rules.initial.robotScanDefault,
                    desiredForce    : 0.0,
                    desiredAngle    : robotPosition[2],
                    desiredLaunch   : false,
                    angleToCenter   : angle180,
                    scanner: {
                        projectile  : [],
                        robot       : [],
                        wall        : [],
                        ALL         : []
                    },
                    projectile: {
                        last_frame  : false,
                        count       : 0,
                        hits        : 0,
                        misses      : 0
                    },
                    proximity       : []
                });
            });

        let teams = new Set();
        robots.forEach(robot => {
            // workaround, setting angle in options caused malformed svg
            // only apply random angle if more than 1 robot and no preset
            if (robot.desiredAngle !== false) {
                Matter.Body.setAngle(robot, robot.desiredAngle);
                robot.angleToCenter = radians_to_center(robot);
            }
            robot.desiredAngle = false;
            teams.add(robot.entityTeam);
        });

        if (Array.from(teams).length == 1) {
            settings.sim.ignoreSingleTeamEndsRound = true;
        } else {
            settings.sim.ignoreSingleTeamEndsRound = false;
        }

        console.log(
            "[sim:" + settings.sim.IID + "::setup_robots]",
            "setup",
                robots.length, "robots",
                Array.from(teams).length, "teams");

        return robots;
    };

    const stop_all_pending_events = () => {
        toggle_pause(true, "reset");

        Loops.stop(settings.sim.IID + "|physics");
        Loops.stop(settings.sim.IID + "|updates");

        clearTimeout(Timeouts.recovery);
        clearTimeout(Timeouts.autorestart);
    }

    const simulation_reset = () => {
        stop_all_pending_events();

        // remove any pre-existing entity references
        for (const type of ["walls", "robots", "projectiles", "sensors"]) {
            Entities[type].forEach((body) => {
                Matter.World.remove(Engine.world, body);
            });
            Entities[type] = [];
        }
        Object.keys(BodyById).forEach((key)=>{
            delete BodyById[key];
        });
        Object.keys(ScanByRobot).forEach((key)=>{
            delete ScanByRobot[key];
        });

        if (Engine) {
            Engine.events = {};
            Matter.World.clear(Engine.world);
            Matter.Engine.clear(Engine);
            Engine = false;

            console.log(
                "[sim:" + settings.sim.IID + "::simulation_reset]",
                "engine disposed");
        }

        seed_prng();
    }

    const simulation_stop = () => {
        console.log(
            "[sim:" + settings.sim.IID + "::simulation_stop]");

        simulation_reset();

        broadcast("simulation stop");
    }

    function simulation_init(paused) {
        if (!state.started) state.started = new Date();

        simulation_reset();

        // ensure simulation starts in paused state
        toggle_pause(true, "init");

        // update client settings
        state.arena = settings.arena;
        state.realtime = settings.sim.realtime;

        // setup all bodies

        Entities.walls = [
            arena_wall(
                settings.arena.width / 2,
                -(settings.arena.wall/2),
                settings.arena.width,
                settings.arena.wall, {
                    entitySubtype: "north"
                }),
            arena_wall(
                -(settings.arena.wall/2),
                settings.arena.height / 2,
                settings.arena.wall,
                settings.arena.height, {
                    entitySubtype: "west"
                }),
            arena_wall(
                settings.arena.width + (settings.arena.wall/2),
                settings.arena.height / 2,
                settings.arena.wall,
                settings.arena.height, {
                    entitySubtype: "east"
                }),
            arena_wall(
                settings.arena.width / 2,
                settings.arena.height + (settings.arena.wall/2),
                settings.arena.width,
                settings.arena.wall, {
                    entitySubtype: "south"
                })
        ]

        Entities.robots = setup_robots();

        Entities.projectiles = [];

        // initialise matter.js physics engine
        Engine = Matter.Engine.create();
        Engine.world.gravity.y = 0;

        // initialise all bodies and generate lookup
        let bodies = Object.values(Entities).flat();
        Matter.World.add(Engine.world, bodies);
        bodies.forEach(body => BodyById[body.id] = body);

        console.log(
            "[sim:" + settings.sim.IID + "::simulation_init]",
            "body count", Object.keys(BodyById).length);

        // initialise collision handlers
        // for both: scanner "hits" and physical collisions

        Matter.Events.on(Engine, 'collisionStart', function(event) {
            detect_projectile_hit(event.pairs);
            detect_wall_collision(event.pairs);
            detect_robot_collision(event.pairs);

            detect_scanner_hit(event.pairs);

            highlight_dynamic_collision(event.pairs, 'cyan');
        });

        Matter.Events.on(Engine, 'collisionActive', function(event) {
            // highlight_dynamic_collision(event.pairs, 'blue');
        });

        Matter.Events.on(Engine, 'collisionEnd', function(event) {
            highlight_dynamic_collision(event.pairs, false); // reset
        });

        // update round state
        state.count.rounds++;
        state.round.end = settings.rules.limit.round;
        state.round.remaining = 0;
        state.round.tsEnded = false;
        state.round.results = false;

        if (typeof paused === "undefined") paused = false;

        simulation_start(paused);
    }

    async function simulation_tick() {
        count_population();

        // state.paused will prevent roundHasEnded check
        let tickForward = !state.paused && !roundHasEnded();

        if (tickForward) {
            var options = {
                frames: state.count.frames,
                framesReloadBrain: settings.sim.physicsFPS * 5,
            };

            if (DEBUG_INJECT_INTO_ROBOT) options['debug'] = debug;

            var [ALL, bots_executor_err] = await brains.async_run_all(
                Entities.robots, options);

            if (!ALL) {
                if (bots_executor_err.length > 0) {
                    for (var i=0; i<bots_executor_err.length;i++) {
                        take_damage(
                            bots_executor_err[i],
                            settings.rules.damage.executorError);
                    }
                } else {
                    toggle_pause(true, "recovery, wait", true);
                    Timeouts.recovery = setTimeout(() => {
                        toggle_pause(false, "recovery, check", true);
                    }, 3000);
                    return;
                }
            }

            scanner_reset();

            state.count.frames++;
            state.count.seconds = state.count.frames / settings.sim.physicsFPS;

            try {
                Matter.Engine.update(Engine, settings.sim.physicsRate);
            } catch (e) {
                console.log(
                    "[sim:" + settings.sim.IID + "::simulation_start]",
                    "MATTERJS", e);
                simulation_stop();
                return false;
            }

            process_rotations();
            process_propulsion();
            process_projectiles();

            update_all_scans();
            update_proximity_alarms();

            cleanup_projectiles();
            cleanup_robots();
        };

        return tickForward;
    }

    async function status_updates() {
        // state is always updated - regardless of run state
        // this ensures that all clients are updated with latest state
        broadcast("update state", {
            robots: Entities.robots.map(toRobots),
            walls: Entities.walls.map(toGenericVertexes),
            projectiles: Entities.projectiles.map(toCircles),
            sensors: Entities.sensors.map(toScanRanges),
            status: state
        });
    }

    async function simulation_start(paused) {
        console.log(
            "[sim:" + settings.sim.IID + "::simulation_start]",
            "reset client states");

        broadcast("reset state");

        state.count.frames = 0;
        state.count.seconds = 0;

        console.log(
            "[sim:" + settings.sim.IID + "::simulation_start]",
            "start client updates...");

        toggle_pause(paused, "default state");

        // reset state
        brains.reset();

        if (settings.sim.realtime) {

            Loops.start(
                settings.sim.IID + "|updates",
                status_updates,
                settings.sim.updateRate);

            Loops.start(
                settings.sim.IID + "|physics",
                simulation_tick,
                settings.sim.physicsRate);

        } else {
            // resolve as soon as possible
            // dev: fastest resolution

            console.time("offlineResolve");

            var running = false,
                lastupdated = false;

            do {
                var now = (new Date).getTime();
                running = await simulation_tick();
                if (lastupdated === false || now - lastupdated > 2000) {
                    console.log(
                        "[sim:" + settings.sim.IID + "::simulation_start]"
                        + "[offline]",
                        "frame:", state.count.frames,
                        "brain runs:", brains.state.runs);
                    lastupdated = now;
                }
            } while(running);

            console.timeEnd("offlineResolve");

            Loops.start(
                settings.sim.IID + "|updates",
                status_updates,
                1000/1);
        };

    }

    function toggle_pause(force_state, reason, nolog) {
        if (!reason) reason = "none";
        if (typeof force_state === "undefined") {
            state.paused = !state.paused;

            if (!nolog) console.log(
                "[sim:" + settings.sim.IID + "::toggle_pause]",
                "toggled, state:", state.paused,
                "reason:", reason);

        } else if (state.paused != force_state) {
            state.paused = force_state;

            if (!nolog) console.log(
                "[sim:" + settings.sim.IID + "::toggle_pause]",
                "state:", state.paused,
                "reason:", reason);
        }
    }

    /*
    simulation loop functionality
    */

    function count_population() {
        let activeTeams = new Set(),
            activeRobots = Entities.robots.filter(r => r.active);

        activeRobots.forEach(r => { activeTeams.add(r.entityTeam) });

        state.count.robots = activeRobots.length;
        state.count.teams = activeTeams.size;
    }

    function roundHasEnded() {
        end_round = [
            state.count.robots,
            (settings.sim.ignoreSingleTeamEndsRound ?
                9999 : state.count.teams)
        ].some(x => x <= 1);

        if (end_round) {
            state.round.end = Math.min(
                state.count.seconds + settings.rules.limit.lastrobot,
                state.round.end);
        }

        // resolve the round earlier when possible
        state.round.remaining = Math.max(0.00,
            state.round.end - state.count.seconds);

        if (state.round.remaining > 0.00) return false;

        // end the round

        toggle_pause(true, "end of round"); // force simulation to pause

        if (determine_round_results()) {
            var stop_simulation = false;

            if (!stop_simulation) {
                // exceeded maximum rounds?
                stop_simulation = (settings.sim.maxRounds > 0
                    && state.count.rounds >= settings.sim.maxRounds);
                console.log(
                    "[sim:" + settings.sim.IID + "::roundHasEnded]",
                    "rounds", state.count.rounds, settings.sim.maxRounds,
                    "stop:", stop_simulation);
            }

            if (!stop_simulation) {
                // exceeded maximum runtime?
                var runtime = ((new Date).getTime() - settings.sim.created.getTime())/1000;
                stop_simulation = (settings.sim.maxRuntime > 0
                    && runtime >= settings.sim.maxRuntime);
                console.log(
                    "[sim:" + settings.sim.IID + "::roundHasEnded]",
                    "runtime", runtime, settings.sim.maxRuntime,
                    "stop:", stop_simulation);
            }

            if (!stop_simulation) {
                // autorestart if required
                var autorestart = settings.sim.autoRestart;
                if (autorestart) {

                    console.log(
                        "[sim:" + settings.sim.IID + "::roundHasEnded]",
                        "new round in", autorestart);

                    Timeouts.autorestart = setTimeout(function(e){
                        simulation_init();
                    }, autorestart * 1000);
                }
            }

            if (stop_simulation) {
                state.stopped = true;
            }
        }

        return true;
    }

    function update_all_scans() {
        Entities.sensors.forEach((sensor, i) => {
            if (!sensor) return;
            Matter.World.remove(Engine.world, sensor);
            Entities.sensors[i] = false;
        });
        Entities.sensors = Entities.sensors.filter(Boolean);

        Entities.robots.forEach(robot => {
            scan_ahead(robot);
        });
    }

    function update_proximity_alarms() {
        Entities.robots.forEach(robot => {
            var shortrange = [];

            shortrange.push(
                ...Entities.walls.map(
                    wall => dict_otherbody(wall, robot)));

            shortrange.push(
                ...Entities.robots.reduce((acc, r) => {
                    if (r.id != robot.id) {
                        acc.push(dict_otherbody(r, robot));
                    };
                    return acc;
                }, []));

            robot.proximity = shortrange
                .filter(x => x.range < (robot.maxDimension * 1.5))
                .sort((a, b) => a.range - b.range);
    });
    }

    function process_rotations() {
        Entities.robots.forEach(robot => {
            if (!robot.active) return;
            if (robot.desiredAngle === false) return;

            var rounding = 4,
                exp = Math.pow(10, rounding),
                angle = utils.clamped_angle(robot.angle, rounding),
                desiredAngle = utils.clamped_angle(robot.desiredAngle, rounding),
                clockwise = (angle - desiredAngle + angle360) % angle360 > Math.PI,
                diff = Math.round(Math.abs(angle - desiredAngle) * exp) / exp,
                rotation = false;

            if (diff < robot.maxRotate) {
                robot.desiredAngle = false;
                rotation = (clockwise ? diff : -diff);
                // console.log("partial", desiredAngle, angle, rotation);
            } else {
                rotation = (clockwise ? robot.maxRotate : -robot.maxRotate);
                // console.log("full", desiredAngle, angle, rotation);
            }
            Matter.Body.rotate(robot, rotation);

            robot.angleToCenter = radians_to_center(robot);
        });
    }

    function process_propulsion() {
        Entities.robots.forEach(robot => {
            if (!robot.active) return;
            if (robot.desiredForce == 0.0) return;

            // cap to 1x +ve, 0.5x -ve
            robot.desiredForce = Math.max(
                Math.min(robot.maxForce, robot.desiredForce),
                -robot.maxForce/2);

            Matter.Body.applyForce(robot, robot.position, {
                x: Math.cos(robot.angle) * robot.desiredForce,
                y: Math.sin(robot.angle) * robot.desiredForce
            });

            robot.desiredForce = 0.0;
        });
    }

    function process_projectiles() {
        Entities.robots.forEach(robot => {
            if (!robot.active) return;
            if (!robot.desiredLaunch) return;

            launch_projectile(robot);
        });
    }

    function cleanup_projectiles() {
        Entities.projectiles.forEach((projectile, i) => {
            if (!projectile) return; // removed during collision detection

            var x = projectile.position.x,
                y = projectile.position.y,
                speed = projectile.speed,
                angularSpeed = projectile.angularSpeed;
                isInvalid = [
                    // out of bounds
                    x < 0,
                    y < 0,
                    x > settings.arena.width,
                    y > settings.arena.height,
                    // slow, dev: start frame is always 0!
                    (speed < settings.rules.physics.projectileMinSpeed
                        && projectile.launchFrame != state.count.frames),
                    // spinning - probably from glancing bounce
                    angularSpeed > 0.0
                ].some(x => Boolean(x));

            if (!isInvalid) return;

            // increment source robot miss stats
            if (projectile.refidSource in BodyById)
                BodyById[projectile.refidSource].projectile.misses++;

            // remove projectile from world, engine
            delete BodyById[projectile.id];
            Matter.World.remove(Engine.world, projectile);
            Entities.projectiles[i] = false;
        });

        // remove blanks from tracking array
        Entities.projectiles = Entities.projectiles.filter(Boolean);
    }

    function cleanup_robots() {
        Entities.robots.forEach(robot => {
            if (!robot.active) return;

            var x = robot.position.x,
                y = robot.position.y;

            if (x < 0 || y < 0 ||
                    x > settings.arena.width || y > settings.arena.height) {

                console.log(
                    "[sim:" + settings.sim.IID + "::cleanup_robots]",
                    "robot exited arena");

                take_damage(robot, robot.maxHealth);
            }
        })
    }

    function determine_round_results() {
        if(state.round.results) {
            console.log(
                "[sim:" + settings.sim.IID + "::determine_round_results]",
                "round results already stored");

            return false;
        }

        // list of robots by time, health, accuracy asc

        let robot_results = Entities.robots
            // map as
            //  [0]name, [1]team, [2]id, [3]lifetime, [4]hits, [5]misses,
            //  [6]health, [7] accuracy, [8]external id
            .map(r => [
                r.entityName, r.entityTeam, r.id,
                r.tkoAt == -1 ? state.count.seconds : r.tkoAt,
                r.projectile.hits,
                r.projectile.misses,
                Math.max(0, 100 - Math.floor(r.damage / r.maxHealth * 100)),
                (r.projectile.hits == 0 && r.projectile.misses == 0 ? 0
                    : r.projectile.hits / (r.projectile.hits + r.projectile.misses)) * 100,
                // dev: variable passtru via results
                r.entityID
            ])
            /* sort: best is last item, example input:
            [
                ["Robot A", "Team A", "1",   30, 7,0,  20, 100],
                ["Robot B", "Team A", "2",   15, 0,2,  50,   0],
                ["Robot C", "Team B", "3",   40, 0,0,   1,   0],
                ["Robot D", "Team B", "4",   30, 9,2,  60,  81.81818181818183],
                ["Robot E", "Team C", "5",   30, 8,3,  40,  72.72727272727273],
                ["Robot F", "Team C", "6",  100, 3,7,  60,  30],
                ["Robot G", "Team D", "7",  100, 7,3,  60,  70],
                ["Robot H", "Team D", "8",  100, 3,2,   1,  60],
                ["Robot I", "Team E", "9",   30, 5,2,  20,  71.42857142857143],
                ["Robot J", "Team E", "10",  30, 0,0,  10,   0],
            ]
            */
            .sort((a, b) =>
                    a[3] > b[3] ?  1    // a longer-lasting
                 : (a[3] < b[3] ? -1    // b longer-lasting
                 : (a[6] > b[6] ?  1    // a healthier
                 : (a[6] < b[6] ? -1    // b healthier
                 : (a[7] > b[7] ?  1    // a more accurate
                 : -1))))               // b more accurate (or same)
            ); // end .sort (robots)

        // list of teams by max time, average health asc
        // dev: average health avoids issues with uneven team participants

        let team_results = Object.entries(robot_results.reduce(
            (acc, robot) => {
                let rname   = robot[0],
                    rteam   = robot[1],
                    rid     = robot[2],
                    rtime   = robot[3],
                    rhealth = robot[6],
                    racc    = robot[7],
                    // create/merge successive rows into teams metadata
                    // [0]min time, [1]max time, [2]robot list, [3]sum health, [4]sum accuracy
                    prev    = acc[rteam] || [ false, false, [], 0, 0];
                prev[2].push([rname, rid, rtime, rhealth, racc]);
                acc[rteam] = [
                    Math.min(prev[0] || state.count.frames, rtime),
                    Math.max(prev[1] || -1, rtime),
                    prev[2],
                    prev[3] + rhealth,
                    prev[4] + racc];
                return acc;
            }, {}))
                /* flatten as [0]name, [1]min time, [2]max time, [3]robots, [4]avg health [5]avg health, example input:
                {
                    "Team A": [ 15,  30, [["Robot B","2",15,  50], ["Robot A","1",30,  20]],  70, 100],
                    "Team E": [ 30,  30, [["Robot J","10",30, 10], ["Robot I","9",30,  20]],  30,  71.42857142857143],
                    "Team C": [ 30, 100, [["Robot E","5",30,  40], ["Robot F","6",100, 60]], 100, 102.72727272727273],
                    "Team B": [ 30,  40, [["Robot D","4",30,  60], ["Robot C","3",40,   1]],  61,  81.81818181818183],
                    "Team D": [100, 100, [["Robot H","8",100,  1], ["Robot G","7",100,  1]],   2, 130],
                }
                */
                .map(x => [
                    x[0],
                    x[1][0],
                    x[1][1],
                    x[1][2],
                    x[1][3] / x[1][2].length,
                    x[1][4] / x[1][2].length])
                /* sort: best is last item, example input:
                [
                    ["Team A",  15,  30, [["Robot B",  "2",  15, 50], ["Robot A", "1",  30, 20]], 35,   50],
                    ["Team E",  30,  30, [["Robot J", "10",  30, 10], ["Robot I", "9",  30, 20]], 15,   35.714285714285715],
                    ["Team C",  30, 100, [["Robot E",  "5",  30, 40], ["Robot F", "6", 100, 60]], 50,   51.36363636363637],
                    ["Team B",  30,  40, [["Robot D",  "4",  30, 60], ["Robot C", "3",  40,  1]], 30.5, 40.909090909090914],
                    ["Team D", 100, 100, [["Robot H",  "8", 100,  1], ["Robot G", "7", 100,  1]],  1,   65],
                ]
                */
                .sort((a, b) =>
                        a[2] > b[2] ?  1    // a longer-lasting
                     : (a[2] < b[2] ? -1    // b longer-lasting
                     : (a[4] > b[4] ?  1    // a healthier
                     : (a[4] < b[4] ? -1    // b healthier
                     : (a[5] > b[5] ?  1    // a more accurate
                     : -1))))               // b more accurate (or same)
                ); // end .sort (teams)

        // win state by team, robots: placed, tie, none
        state.round.results = {
            "teams": determine_placement(team_results, [2,4,5]),
            "robots": determine_placement(robot_results, [3,6,7])
        };

        console.log(
            "[sim:" + settings.sim.IID + "::determine_round_results]"
            + "[round " + state.count.rounds + " results] robots",
            state.round.results.robots
                .map((x, idx) =>
                    [idx + 1, x.participants.map(y => y[0]).join(", ")])
        );

        // get an ordered list of robot names for sim instance
        // dev: used to ensure csv header sequence

        var robotnames = []
        Entities.robots.forEach(robot => robotnames.push(robot.entityName));
        robotnames.sort();

        if (simlog_id) {
            simlogger.write(
                simlog_id,
                state.count.rounds,
                robotnames,
                state.round.results.robots);

            simlogger.write_json(
                simlog_id,
                state.count.rounds,
                state.round.results.robots);
        }

        console.log(
            "[sim:" + settings.sim.IID + "::determine_round_results]"
            + "[round " + state.count.rounds + " results] teams",
            state.round.results.teams
                .map((x, idx) =>
                    [idx + 1, x.participants.map(y => y[0]).join(", ")])
        );

        return true;
    }

    function determine_placement(scoreList, positions) {
        // returns participants list indexed by placement from 1st to last
        // scoreList presorted by determine_round_results()

        let places = [],
            placement = -1,
            comp_next = false,
            comp_prev = false,
            next = false;

        // consume until list is exhausted
        while(scoreList.length > 0) {
            next = scoreList.pop();

            // compares properties of scoreList items to increment placement
            comp_next = "";
            positions.forEach(idx => comp_next += "[" + next[idx] + "]");
            if (comp_prev === false || comp_next != comp_prev) placement++;
            comp_prev = comp_next;

            if (placement >= places.length) {
                places.push({"decision": "none", "participants": []});
            }

            places[placement].participants.push(next);

            switch(places[placement].participants.length) {
                case 1: places[placement].decision = "placed"; break;
                default: places[placement].decision = "tie"; break;
            }
        };

        return places;
    }

    /*
    scanner detection and handling
    */

    function scanner_reset() {
        ScanByRobot = {};
        Entities.robots.forEach(robot => {
            robot.scanner = {
                "ALL": [],
                "wall": [],
                "robot": [],
                "projectile": []
            }
        });
    }

    function scanner_store(robotid, bodyid) {
        // called when scan cone detects a body (collision)
        if (!(robotid in ScanByRobot)) ScanByRobot[robotid] = {};
        if (!(bodyid in ScanByRobot[robotid])) {
            if (bodyid in BodyById) {
                ScanByRobot[robotid][bodyid] = BodyById[bodyid];
            }
        }
    }

    function dict_otherbody(body, robot) {
        let entityType = body.entityType,
            dx = body.position.x - robot.position.x, // (from robot.x)
            dy = robot.position.y - body.position.y, // (from robot.y)
            dict = {
                entityType      : entityType,
                id              : body.id,
                x               : body.position.x,
                y               : body.position.y,
                speed           : Math.round(body.speed * 1000)/1000,
                angle           : utils.clamped_angle(body.angle),
                // absolute angle from robot to body
                angleFromRobot  : utils.clamped_angle(Math.atan2(-dy, dx)),
                range           : Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2))
            };

        if (dict.entityType == "robot") {
            dict.active = body.active;
            dict.name = body.entityName;
            dict.team = body.entityTeam;
        } else if (dict.entityType == "wall") {
            dict.angle = 0;
            dict.entitySubtype = body.entitySubtype;
            // dev: offset wall thickness
            var walloffset = (settings.arena.wall / 2);
            // perpendicular to wall
            switch (body.entitySubtype) {
                case "west":
                    dict.range = Math.abs(dx + walloffset);
                    dict.angleFromRobot =  utils.clamped_angle(angleWest);
                    break;
                case "east":
                    dict.range = Math.abs(dx - walloffset);
                    dict.angleFromRobot =  utils.clamped_angle(angleEast);
                    break;
                case "north":
                    dict.range = Math.abs(dy + walloffset);
                    dict.angleFromRobot =  utils.clamped_angle(angleNorth);
                    break;
                case "south":
                    dict.range = Math.abs(dy - walloffset);
                    dict.angleFromRobot =  utils.clamped_angle(angleSouth);
                    break;
                default:
                    throw "invalid wall";
            }
        }

        return dict;
    }

    function scanner_update() {
        // updates all robot scanners
        Object.keys(ScanByRobot).forEach(robotid => {
            let robot = BodyById[robotid],
                buffer = {
                    "wall": [],
                    "robot": [],
                    "projectile": []
                };

            if (!robot) return; // in event of reset

            // optimisation: raytrace walls for reference later
            let raytracedWalls = utils.closest_edges(
                robot.position.x, robot.position.y,
                utils.clamped_angle(robot.angle),
                [0, settings.arena.height, settings.arena.width, 0]);

            // iterate through each detected body
            var detections = ScanByRobot[robotid];
            Object.keys(detections).forEach(bodyid => {
                let detected = dict_otherbody(detections[bodyid], robot);

                // special: if wall, use raytraced range instead of perpendicular
                if (detected.entityType == "wall") {
                    let raytracedWall = raytracedWalls
                        .find(x => x[0] == detected.entitySubtype);
                    if (raytracedWall) {
                        // replace properties with raytraced version
                        detected.range = raytracedWall[1];
                        detected.angleFromRobot = robot.angle;
                    } else {
                        // not found, happens if robot angled away
                        // from the wall, but the scanner cone is still in
                        // contact. this tends to happens when robot is
                        // rapidly turning or when bouncing from edges
                        detected = false;
                    }
                }

                // allow skipping entities
                if (detected === false) return;

                buffer[detected.entityType].push(detected);
            });

            // generate ALL - separated so we can preprocess walls
            Object.keys(buffer).forEach(k => {
                buffer.ALL = (buffer.ALL || []).concat(buffer[k]);
            })

            // sort all detection keys by nearest on top
            Object.keys(buffer).forEach(k => {
                buffer[k].sort((a, b) => a.range - b.range);
            });

            robot.scanner = buffer;
        })
    }

    function detect_scanner_hit(pairs) {
        pairs.forEach(pair => {
            var valid = only_pair(pair, "scan", "wall");
            if (valid) {
                scanner_store(valid[0].refidRobot, valid[1].id);
                return;
            }

            var valid = only_pair(pair, "scan", "robot");
            if (valid && valid[0].refidRobot != valid[1].id) {
                // other robots only - can't detect self
                scanner_store(valid[0].refidRobot, valid[1].id);
                return;
            }

            var valid = only_pair(pair, "scan", "projectile");
            if (valid && valid[0].refidRobot != valid[1].refidSource) {
                // other robot projectiles only
                scanner_store(valid[0].refidRobot, valid[1].id);
                return;
            }
        });

        scanner_update();
    }

    /*
    collision detection and handling
    */

    function highlight_dynamic_collision(pairs, hex) {
        pairs.forEach(pair => {
            var valid = only_pair(pair, "robot", "wall");
            if (valid && valid[0].active) {
                valid[0].render.fillStyle = hex || damage_rgb(valid[0]);
                return;
            }

            var valid = only_pair(pair, "robot", "projectile");
            if (valid && valid[0].active) {
                valid[0].render.fillStyle = hex || damage_rgb(valid[0]);
                return;
            }

            var valid = only_pair(pair, "robot", "robot");
            if (valid) {
                if (valid[0].active)
                    valid[0].render.fillStyle = hex || damage_rgb(valid[0]);

                if (valid[1].active)
                    valid[1].render.fillStyle = hex || damage_rgb(valid[1]);

                return;
            }

        });
    }

    function detect_projectile_hit(pairs) {
        pairs.forEach(pair => {
            var valid = only_pair(pair, "robot", "projectile");
            if (!valid) return;

            take_damage(valid[0], settings.rules.damage.projectile);

            // increment source robot hit stats
            if (valid[1].refidSource in BodyById)
                BodyById[valid[1].refidSource].projectile.hits++;

            // remove projectile from world, engine
            delete BodyById[valid[1].id];
            for (var i=0; i<Entities.projectiles.length; i++) {
                if (valid[1] == Entities.projectiles[i]) {
                    Entities.projectiles[i] = false;
                    break;
                }
            }
            Matter.World.remove(Engine.world, valid[1]);
        });
    }

    function detect_wall_collision(pairs) {
        pairs.forEach(pair => {
            var valid = only_pair(pair, "robot", "wall");
            if (!valid) return;

            take_damage(valid[0], settings.rules.damage.wall);
        });
    }

    function detect_robot_collision(pairs) {
        pairs.forEach(pair => {
            valid = only_pair(pair, "robot", "robot");
            if (!valid) return;

            take_damage(valid[0], settings.rules.damage.robot);
            take_damage(valid[1], settings.rules.damage.robot);
        });
    }

    /*
    single-body methods
    all act on a single physics body
    */

    function scan_ahead(robot) {
        if (!robot.active) return;

        var lenMax = settings.rules.limit.scanLenMax,
            lenMin = settings.rules.limit.scanLenMin,
            scaleMax = settings.rules.limit.scanScaleMax,
            scaleMin = settings.rules.limit.scanScaleMin,
            scanWidthMultiplier = Math.min(
                Math.max(lenMax/robot.desiredScan, scaleMin),
                scaleMax),
            height = Math.min(
                Math.max(robot.desiredScan, lenMin),
                lenMax),
            width = robot.maxDimension * scanWidthMultiplier; // far edge width
            ws = robot.maxDimension, // near edge width
            scanSlope = 1 - ws/width, // calc slope of trapezoid
            cx = (height / 3) * ((2*width + ws) / (width + ws)), // centroid
            center_offset = cx + (robot.maxDimension / 2);

        var scanner = Matter.Bodies.trapezoid(
            robot.position.x + center_offset * Math.cos(robot.angle),
            robot.position.y + center_offset * Math.sin(robot.angle),
            width, height, scanSlope, {
                // physics properties (matterjs)
                isSensor        : true,
                angle           : robot.angle - angle90,
                render: {
                    // dev: might be ignored by client-side renderer
                    fillStyle   : "rgba(150, 150, 255, 0.1)"
                },
                collisionFilter: {
                    category    : collision_targets
                },
                // custom properties
                entityType      : "scan",
                refidRobot      : robot.id,
                added           : (new Date()).getTime(),
                scanDistance    : height // min-max clamped
            }
        );

        Matter.World.addBody(Engine.world, scanner);
        Entities.sensors.push(scanner);
    }

    function launch_projectile(body) {
        // cap to server frames
        var diff = state.count.frames - (body.projectile.last_frame || 0);
        if (diff < settings.rules.physics.projectileRate) return false;
        body.projectile.last_frame = state.count.frames;

        body.projectile.count++;

        var projectile = Matter.Bodies.circle(
            body.position.x +
                ((body.maxDimension + settings.sim.projectileLeadOffset)
                    * Math.cos(body.angle)),
            body.position.y +
                ((body.maxDimension + settings.sim.projectileLeadOffset)
                    * Math.sin(body.angle)),
            settings.sim.projectileSize, {
                // physics properties (matterjs)
                inertia         : Infinity,
                restitution     : 0,
                frictionAir     : 0,
                friction        : 0,
                frictionStatic  : 1,
                collisionFilter: {
                    mask        : collision_targets
                },
                // custom properties
                launchFrame     : state.count.frames,
                entityType      : "projectile",
                refidSource     : body.id
            }
        );

        Matter.World.addBody(Engine.world, projectile);
        Entities.projectiles.push(projectile);
        BodyById[projectile.id] = projectile;

        Matter.Body.applyForce(projectile, projectile.position, {
            x: Math.cos(body.angle) * settings.rules.physics.projectileForce,
            y: Math.sin(body.angle) * settings.rules.physics.projectileForce
        });

        // apply recoil when firing projectile
        var recoilForce = -Math.abs(settings.rules.physics.projectileRecoilForce);
        if (recoilForce) {
            Matter.Body.applyForce(body, body.position, {
                x: Math.cos(body.angle) * recoilForce,
                y: Math.sin(body.angle) * recoilForce
            });
        }

        if ('desiredLaunch' in body) body.desiredLaunch = false;
    }

    function take_damage(body, intensity) {
        if (typeof intensity === "undefined") intensity = 1;

        // works on geometric body and robot
        let robot = false;
        if (body.entityType == "robot") {
            robot = body;
        } else {
            robot = body.parent;
        }

        robot.damage += intensity;
        if (robot.damage >= robot.maxHealth) {
            if (robot.active) knockout_robot(robot)
        }
    }

    function knockout_robot(robot) {
        robot.active = false;
        robot.tkoAt = state.count.seconds;
        broadcast("robot knockout", {
            "id": robot.id,
            "x": robot.position.x,
            "y": robot.position.y
        });
    }

    /* debug behaviours and methods */

    function debug_all_spread(coordinates, force) {
        // apply outward force to all robots from given coordinates
        // affects: inactive, active robots

        if (!force) force = 0.01;

        Entities.robots.forEach(robot => {
            const deltaVector = Matter.Vector.sub(robot.position, coordinates);
            const normalizedDelta = Matter.Vector.normalise(deltaVector);
            const forceVector = Matter.Vector.mult(normalizedDelta, force);
            Matter.Body.applyForce(robot, robot.position, forceVector);
        });
    }

    // public properties/methods

    const callfunc = (func) => {
        return () => {
            if (state.started && settings.sim.locked) {
                console.log(
                    '[sim:' + settings.sim.IID + '::' + func.name + ']',
                    'denied, simulation locked');
                return false;
            }
            func();
        }
    }

    Arena = {
        // read-only
        settings: settings,
        state: state,
        // setters
        set_socket_io: set_socket_io,
        set_socket_room: set_socket_room,
        set_simlog_id: set_simlog_id,
        // methods
        start: callfunc(simulation_init),
        pause: callfunc(toggle_pause),
        reset: callfunc(simulation_reset),
        stop: simulation_stop,
        // methods: debug
        debug_inject_robot: false,
        debug_all_spread: debug_all_spread,
        debug_all_rotate_right: () => {
            Entities.robots.forEach(robot => {
                robot.desiredAngle = utils.clamped_angle(
                    robot.angle + 0.174533);
            });
        },
        debug_all_rotate_left: () => {
            Entities.robots.forEach(robot => {
                robot.desiredAngle = utils.clamped_angle(
                    robot.angle - 0.174533);
            });
        },
        debug_all_move_forward: () => {
            Entities.robots.forEach(robot => {
                robot.desiredForce = robot.maxForce;
            });
        },
        debug_all_face_center: () => {
            Entities.robots.forEach(robot => {
                robot.desiredAngle = radians_to_center(robot);
            });
        },
        debug_all_fire: () => {
            Entities.robots.forEach(robot => {
                robot.desiredLaunch = true;
            });
        }
    }

    return Arena;

}; // Factory()

module.exports = {
    Factory: Factory,
    // global debug
    set_debug_inject_into_robot: set_debug_inject_into_robot
};
