rad_deg = (angle, rounding) => {
    var degrees = angle * 180 / Math.PI;
    if (rounding) {
        rounding = Math.pow(10, rounding);
        degrees = Math.round(degrees * rounding) / rounding;
    }
    return degrees;
}

clamped_angle = (angle, rounding) => {
    // restrict angle (radians) to [0, 6.28319] (360 degrees)
    // IMPORTANT: canvas angle: +ve angle is clockwise
    // https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/rotate

    var maxrads = Math.PI * 2;
    while (angle < 0) angle += maxrads;
    angle = angle % maxrads;
    if (rounding) {
        rounding = Math.pow(10, rounding);
        angle = Math.round(angle * rounding) / rounding;
    }
    return angle;
};

abs_bearing = (angle, rounding) => {
    // returns bearing in degrees
    // 0=NORTH, 90=E, 180=S, 270=W

    angle = clamped_angle(angle);
    degrees = 90 + rad_deg(angle);
    while (degrees < 0) degrees += 360;
    degrees = degrees % 360;
    if (rounding) {
        rounding = Math.pow(10, rounding);
        degrees = Math.round(degrees * rounding) / rounding;
    }
    return degrees;
};

angle_to = (x1, y1, x2, y2) => {
    // angle from x1, y1 to x2, y2

    let dx = x2 - x1,
        dy = y2 - y1,
        rads = Math.atan2(dy, dx);
    return rads;
};

closest_edges = (x, y, angle, bbox) => {
    // distance to intersecting bounding box edges
    // returns list, closest edge first
    // dev: canvas has inverted y-axis: top-left y > bottom-right y

    // https://stackoverflow.com/a/3197924
    var vx = Math.cos(angle),
        vy = Math.sin(angle)
        tests = [
            // top-left
            [(bbox[0]-x)/vx, "west"],
            [(bbox[2]-x)/vx, "east"],
            // bottom-right
            [(bbox[1]-y)/vy, "south"],
            [(bbox[3]-y)/vy, "north"]
        ].filter(t => isFinite(t[0]) && t[0] > 0.0)
         .sort((a, b) => a[0] > b[0] ? 1 : -1);

    if (tests.length == 0) return false;

    return tests.map(t => [t[1], t[0], x + (vx * t[0]), y + (vy * t[0])]);
}

color_gradient = (fadeFraction, rgbColor1, rgbColor2, rgbColor3) => {
    // 2/3 colour gradient for value 0.0 - 1.0
    // adapted from https://stackoverflow.com/a/61396704

    var color1 = rgbColor1;
    var color2 = rgbColor2;
    var fade = fadeFraction;

    if (rgbColor3) {
      fade = fade * 2;

      if (fade >= 1) {
        fade -= 1;
        color1 = rgbColor2;
        color2 = rgbColor3;
      }
    }

    var diffRed = color2.red - color1.red;
    var diffGreen = color2.green - color1.green;
    var diffBlue = color2.blue - color1.blue;

    var gradient = {
      red: parseInt(Math.floor(color1.red + (diffRed * fade)), 10),
      green: parseInt(Math.floor(color1.green + (diffGreen * fade)), 10),
      blue: parseInt(Math.floor(color1.blue + (diffBlue * fade)), 10)
    };

    return 'rgb(' +
        gradient.red + ',' +
        gradient.green + ',' +
        gradient.blue + ')';
};

module.exports = {
    rad_deg: rad_deg,
    clamped_angle: clamped_angle,
    abs_bearing: abs_bearing,
    angle_to: angle_to,
    closest_edges: closest_edges,
    color_gradient: color_gradient
};

/*
// verify: abs_bearing output
// dev: +ve radians = clockwise

for (const rad of [
    0,              // 0 degrees, 90 bearing
    Math.PI/4,      // 45 degrees, 45 bearing
    Math.PI/3,      // 60 degrees, 30 bearing
    Math.PI/2,      // 90 degrees, 0 bearing
    Math.PI,        // 180 degrees, 270 bearing
    Math.PI*3/2,    // 270 degrees, 180 bearing
    Math.PI*2,      // 360 degrees, 90 bearing
    -1.965000469985275
]) {
console.log(
    "angle (rad)",
    clamped_angle(rad,2),
    "degrees",
    rad_deg(rad,2),
    "heading",
    abs_bearing(rad,2));
}

xxxx

// */
