const fs = require('fs');

const utils = require('./utils.js');

const arenaman = require('./arenaman.js');
const robotman = require('./robotman.js');

const simlogger = require('./battle/simlogger.js');

const { resolve } = require("path");

const dircomp = resolve("./static/compstates/");

let Queue = [];

let BalancerPorts = {};

const set_balancer_ports = (from_parent) => {
    BalancerPorts = from_parent;
}

const get_path = (compid) => {
    pathcomp = resolve(dircomp + '/comp' + compid + '.json');
    if (!pathcomp.startsWith(dircomp)) throw Error("invalid path: " + pathcomp);
    return pathcomp;
}

const competitionVersus = (compid, robotlist, matchOptions, arenaOptions) => {
    compid = utils.sanitise_string(compid);

    let pathcomp = get_path(compid);

    let Competition = {};

    if (fs.existsSync(pathcomp)) {
        Competition = JSON.parse(fs.readFileSync(pathcomp));
        Competition.status = 'resume';

        console.log('[compman] existing');
    } else {
        // dev: may contain duplicates for robots which are reused
        let robotids = robotlist.replaceAll('\n', ',').replaceAll('\r', ',').split(/(?:,| )+/);

        let robotmetadata = robotman.loadlist(robotids, ['id', 'name', 'team']);

        // deduplicated robot ids
        robotids = Object.keys(robotmetadata);

        // sanity check
        let countrobots = robotids.length;
        if (countrobots < 2) {
            console.log('[compman] less than 2 robots');
        }

        // shuffle robots order
        robotids.sort(() => Math.random() - 0.5);

        // build groups

        let Groups = {},
            gid = -1;

        robotids.forEach((robotid) => {
            if (gid in Groups && Groups[gid].length < matchOptions.groupMax) {
                Groups[gid].push(robotid)
            } else {
                gid++;
                Groups[gid] = [robotid];
            }
        });

        // build fair matches: GvsG

        let GvsG = {},
            countgroups = Object.values(Groups).length;

        if (countgroups > 1) {
            // remap each group to each other
            Object.entries(Groups).forEach(([idx1, robots1]) => {
                Object.entries(Groups).forEach(([idx2, robots2]) => {
                    if (idx1 == idx2) return true;
                    var key = [
                        utils.num2letters(idx1),
                        utils.num2letters(idx2)].sort().join("-");
                    if (key in GvsG) return true;
                    GvsG[key] = [...robots1, ...robots2];
                });
            });
        } else if (countgroups == 1) {
            // only 1 group and > 1 robot
            GvsG['single'] = Groups[0];
        }

        console.log(
            '[compman]' +
            'robots:', countrobots,
            'groups:', countgroups,
            'matches:', Object.values(GvsG).length,
                GvsG);

        // generate match list and metadata

        let Matches = [],
            arenaids = [],
            lineups = [];

        // shuffle group order
        let shuffled_groups = Object.entries(GvsG).sort(() => Math.random() - 0.5);

        for (const [m_id, m_robotids] of shuffled_groups) {
            let config = {
                'robots': m_robotids,
                // passthru deref
                'matchOptions': JSON.parse(JSON.stringify(matchOptions)),
                'arenaOptions': JSON.parse(JSON.stringify(arenaOptions)),
                'serverOptions': {},
            };

            let arenaid = utils.secure_keystring(32),
                balancerid = utils.balancerid(arenaid);

            config.matchOptions.matchid = m_id;
            config.arenaOptions.arenaid = arenaid;
            config.serverOptions.balancerid = balancerid;

            Matches.push(config);

            // client side data

            arenaids.push(arenaid);

            let m_robotmetadata = [];
            m_robotids.forEach((rid) => m_robotmetadata.push(robotmetadata[rid]));

            lineups.push({
                'arenaid': arenaid,
                'matchid': m_id,
                'robots': m_robotmetadata,
            });
        };

        // persist competition data

        Competition = {
            'compid': compid,
            'completed': false,
            'maxRounds': arenaOptions.maxRounds || 1,
            'robots': Object.values(robotmetadata),
            'arenaids': arenaids,
            'lineups': lineups,
            'Matches': Matches,
        }

        fs.writeFileSync(pathcomp, JSON.stringify(Competition));

        Competition.status = 'new';

        console.log('[compman] new');
    }

    // broadcast via parent

    process.send({
        "type": "broadcast.notify",
        "payload": {
            action: "compman.queueArenas",
            Matches: Competition.Matches,
        }
    });

    return Competition;
}

// cache: reduce hit on disk i/o
let cachedStatus = {};

const competitionStatus = (compid) => {
    compid = utils.sanitise_string(compid);

    // cache: reduce hit on disk i/o
    let now = (new Date()).getTime();
    if (compid in cachedStatus) {
        if (now - cachedStatus[compid][0] < 3000) {
            return cachedStatus[compid][1];
        }
    }

    let pathcomp = get_path(compid);

    if (fs.existsSync(pathcomp)) {
        Competition = JSON.parse(fs.readFileSync(pathcomp));
    } else {
        return false;
    }

    let results = {
        'maxRounds': Competition.maxRounds,
        'byArenas' : {},
        'byRobots' : {},
        'balanceridByArenas': {}
    };

    // prepopulate from initial configuration
    Competition.lineups.forEach(L => {
        results.byArenas[L.arenaid] = L;
        // dev: useful when running local servers
        let balancerid = utils.balancerid(L.arenaid),
            port = BalancerPorts[balancerid] || false;
        results.balanceridByArenas[L.arenaid] = [balancerid, port]
    });

    // iterate through each arena
    Competition.arenaids.forEach((arenaid) => {
        completedRounds = simlogger.read_json(arenaid);
        // initialise extra data in lineups
        results.byArenas[arenaid].rounds = 0;
        if (!completedRounds) return false;
        // mark this lineup as (at least) started
        results.byArenas[arenaid].started = true;
        // build reference lookup for robots in lineup
        // dev: makes it easier for later referencing
        let lookuprobots = {};
        results.byArenas[arenaid].robots.forEach(r => lookuprobots[r.id] = r);
        // iterate through each round
        completedRounds.forEach((round, roundidx) => {
            // prevent extra rounds from affecting ranking
            if (roundidx > results.maxRounds) return;
            // increment round counter
            results.byArenas[arenaid].rounds++;
            // iterate through each placing
            for (const [place, placeMetadata] of Object.entries(round.placements)) {
                // base points: 1st place = 1; 2nd place = 2; etc
                let points = parseInt(place) + 1;
                // penalise ties by number of robots sharing the tie
                if (placeMetadata.decision == 'tied') {
                    points += placeMetadata.participants.length;
                }
                // robots accumulate points
                placeMetadata.participants.forEach(robot => {
                    let rname = robot[0],
                        rteam = robot[1],
                        rid = robot[8];
                    if (rid in results.byRobots) {
                        results.byRobots[rid][0] += points;
                        results.byRobots[rid][4]++;
                    } else {
                        results.byRobots[rid] = [
                            points, rid, rname, rteam, 1];
                    }
                    // accumulate in .byArenas as well
                    try {
                        lookuprobots[rid].score = (lookuprobots[rid].score || 0) + points;
                    } catch(err) {
                        console.log('[compman] lookup failed', arenaid, rid, err);
                    }
                });
            } // forEach (placing)
        }); // forEach (round)
    }) // forEach (arenaid)

    results.byRobots = Object.values(
        results.byRobots).sort((a,b) => a[0] > b[0] ? 1 : -1);

    results.byArenas = Object.values(results.byArenas);

    // cache: reduce hit on disk i/o
    cachedStatus[compid] = [now, results];

    return results;
}

const queueArenas = (Matches) => {
    // received broadcast from parent process

    // index in-memory matches
    let queuedarenas = [];
    Queue.forEach(Match => {
        queuedarenas.push(Match.arenaOptions.arenaid);
    });

    Matches.forEach((Match) => {
        if (Match.serverOptions.balancerid != process.env.BALANCERID) return;

        if (Match.arenaOptions.arenaid in queuedarenas) {
            // state: current match is queued
            // do nothing, queue will proceed as usual
            console.log(
                '[' + process.env.BALANCERID + '::compman]',
                'already queued:', Match.arenaOptions.arenaid,
            );
        } else {
            // state: current match not in queue
            // try to determine current state
            let rounds = simlogger.read_json(Match.arenaOptions.arenaid);
            if (rounds) {
                // state: previous rounds exist
                let completed = rounds.length;
                if (completed == Match.arenaOptions.maxRounds) {
                    // state: ran, and was completed
                    // nothing needs to be done
                    console.log(
                        '[' + process.env.BALANCERID + '::compman]',
                        'queue, completed:', Match.arenaOptions.arenaid,
                    );
                } else {
                    // state: ran, and not complete
                    Match.arenaOptions.maxRounds -= completed;
                    Match.simulation = arenaman.Arena.serverside(
                        Match.robots,
                        Match.arenaOptions,
                    );
                    Queue.push(Match);
                    console.log(
                        '[' + process.env.BALANCERID + '::compman]',
                        'queue, resume:', Match.arenaOptions.arenaid,
                        'left:', Match.arenaOptions.maxRounds,
                    );
                }
            } else {
                // state: entirely new - no previous rounds
                // reserve the arena, but dont start it yet
                // dev: prevents somebody from taking over the arena id
                Match.simulation = arenaman.Arena.serverside(
                    Match.robots,
                    Match.arenaOptions,
                );
                Queue.push(Match);
                console.log(
                    '[' + process.env.BALANCERID + '::compman]',
                    'queue, new:', Match.arenaOptions.arenaid,
                );
            }
        }
    });
}

let logQueueEmpty = false;
let logActiveMatch = false;

setInterval(() => {
    if (Queue.length > 0) {
        logQueueEmpty = false;

        // resolve the queue
        let match = Queue[0];

        if (match.simulation.state.started) {
            if (match.simulation.state.stopped) {
                // sim has finished
                Queue.shift();
                logActiveMatch = false;
            } else {
                if (logActiveMatch) {
                    console.log(
                        '[' + process.env.BALANCERID + '::compman]',
                        'active match:', match.arenaOptions.arenaid)
                }
                // sim still running
                // update states
            }
        } else {
            // new sim
            match.simulation.start();
            logActiveMatch = false;
        }
    } else {
        if (!logQueueEmpty) {
            // one-off log only
            logQueueEmpty = true;
            console.log(
                '[' + process.env.BALANCERID + '::compman]',
                'queue empty'
            );
        }
    }
}, 1000);

module.exports = {
    competitionVersus: competitionVersus,
    queueArenas: queueArenas,
    status: competitionStatus,
    set_balancer_ports: set_balancer_ports,
};
