const numCPUs = require("os").cpus().length;

const { fork } = require('node:child_process');

/* external config */

const dotenv = require('dotenv');

dotenv.config();

let PORT = parseInt(process.env.PORT) || 5000;
let PROCS = Math.min(numCPUs, parseInt(process.env.PROCS) || numCPUs);

console.log("[multi] starting", PROCS, "processes...");

let Processes = {}; // index by port
let Heartbeats = {}; // index by port

let BalancerPorts = {};  // index by balancerid

let SocketmanStates = {}; // index by socketman.js::identifier

/* initilisation and recovery */

const fork_child_process = (listen, balancerid) => {
    console.log('[multi] fork', listen);

    BalancerPorts[balancerid] = listen;

    let proc = fork('server.js', {
        'env': {
            "PORT": listen,
            "MULTINODE": 1,
            "BALANCERID": balancerid,
        },
        'silent': true,
        'detached': false
    });

    // capture and output stdio, stderr

    proc.stdio[1].on('data', (data) => {
        Heartbeats[listen] = new Date();

        console.log(
              "[multi][" + listen + ":stdout]"
            + data.toString().trim());
    });

    proc.stdio[2].on('data', (data) => {
        Heartbeats[listen] = new Date();

        console.log(
              "[multi][" + listen + ":stderr]"
            + data.toString().trim());
    });

    // child -> parent IPC handlers

    proc.on('message', data => {
        Heartbeats[listen] = new Date();

        if (data.type == "socketman.status") {
            SocketmanStates[data.identifier] = [listen, data];
        } else if (data.type == "socketman.deregisterArena") {
            // send only to process with actual arena
            send_by_arenaid(data, data.arenaid);
        } else if (data.type == "broadcast.notify") {
            // client-to-client broadcast, optionally reflect data
            send_to_all(data);
        } else {
            console.log("[multi] unhandled IPC", data.type);
        }
    });

    // recover dead processes

    proc.on('close', () => {
        console.log("[multi] process died", listen);

        delete Processes[listen];

        Processes[listen] = fork_child_process(listen, balancerid);
    });

    return proc;
}

for (var idx=0; idx<PROCS; idx++) {
    ((idx) => {
        const listen = PORT + idx;
        const prockey = '' + listen;
        const balancerid = idx + 1;

        Processes[prockey] = fork_child_process(listen, balancerid);
    })(idx);
};

/* IPC message sending */

const send_by_arenaid = (payload, arenaid) => {
    // iterate through all processes
    var entries = Object.entries(SocketmanStates);
    for (var i=0; i<entries.length; i++) {
        const [_, [listen, Status]] = entries[i];
        // iterate arena list and locate it
        for (var j=0; j<Status.Arenas.length; j++) {
            if (Status.Arenas[j].AID == arenaid) {
                let prockey = '' + listen,
                    P = Processes[prockey];

                if (P && P.send) {
                    P.send(payload)
                } else {
                    console.log(
                        "[multi] unresponsive arena process",
                        prockey)
                }

                break;
            }
        } // end for(Arenas)

        /* continue scanning - in rare situatiuons (load balancer
            change), duplicate arenas might remain using old
            load balancer configuration, therefore send to all
            found instances (esp. deregisterArena) */

    } // end for(Status/Processes)
}

const send_to_all = (payload) => {
    for (const [prockey, P] of Object.entries(Processes)) {
        if (P && P.send) {
            P.send(payload)
        } else {
            console.log("[multi] unresponsive child", prockey)
        }
    };
}

/* state sharing */

const broadcast_status = () => {
    let consolidated = {
        "type": "multiserver.status",
        "Heartbeats": {},
        "Arenas": [],
        "RoboSockets": {},
        "ViewerSockets": {},
        "BalancerPorts": BalancerPorts,
    };

    for (const [identifier, [listen, Status]]
            of Object.entries(SocketmanStates)) {

        // arenas
        Status.Arenas.forEach(arena => {
            arena["source.identifier"] = identifier;
            arena["source.port"] = listen;
        });
        consolidated.Arenas = consolidated.Arenas.concat(Status.Arenas);

        // robot sockets

        for (const [arenaid, Sockets]
                of Object.entries(Status.RoboSockets)) {

            if (arenaid in consolidated.RoboSockets) {
                consolidated.RoboSockets[arenaid] = {
                    ...consolidated.RoboSockets[arenaid], ...Sockets}
                console.log(
                    "[multi] RoboSockets duplicate:",
                    identifier, arenaid);
            } else {
                consolidated.RoboSockets[arenaid] = Sockets;
            }
        }

        // viewer sockets

        for (const [arenaid, Sockets]
                of Object.entries(Status.ViewerSockets)) {

            if (arenaid in consolidated.ViewerSockets) {
                consolidated.ViewerSockets[arenaid] = {
                    ...consolidated.ViewerSockets[arenaid], ...Sockets}
                console.log(
                    "[multi] ViewerSockets duplicate:",
                    identifier, arenaid);
            } else {
                consolidated.ViewerSockets[arenaid] = Sockets;
            }

        }

        // extra: heartbeats
        consolidated.Heartbeats = Heartbeats;
    }

    send_to_all(consolidated);
}

setInterval(broadcast_status, 3000);
