const path = require('path');

// base http server

const express = require("express");
const app = express();
const http = require("http").createServer(app);

app.set('views', path.join(__dirname, 'views'));

// cookie support

const cookieParser = require('cookie-parser')
app.use(cookieParser())

// jinja templating

const expressNunjucks = require('express-nunjucks');

expressNunjucks(app, {watch: true, noCache: true});

/* general utilities */

const utils = require('./utils.js');

/* external config */

const dotenv = require('dotenv');

dotenv.config();

let APIKEY = String(process.env.APIKEY || '').trim();
if (!APIKEY) {
    APIKEY = utils.secure_keystring(32);
    console.log(
        "[WARNING][.env] missing APIKEY, using ", APIKEY);
}

/* http and socket port */

let PORT = parseInt(process.env.PORT || 5001);

http.listen(PORT, () => console.log("listening at", PORT));

/* load balancer */

utils.use_balancer(process.env.PROCS);

/* socket management */

const { Server } = require("socket.io");
const io = new Server(http);

const socketman = require("./socketman.js");

socketman.set_identifier(PORT);

socketman.init(io);

/* competition management */

const compman = require("./compman.js");

/* robot management */

const robotman = require("./robotman.js");

/* simulator */

const { set_debug_inject_into_robot } = require("./battle/simulation.js");
const { check } = require("./battle/simlogger.js");

set_debug_inject_into_robot(process.env.SIMULATOR_INJECT_DEBUG_INTO_ROBOT);

/* expressjs */

app.use(express.static("static"));
app.use(express.json());
app.use(express.urlencoded({ extended: true }));


// routes: robot

app.get('/robot/samples', (req, res) => {
    res.render('robot.samples.html', {
        'config': {
            'cdn': 'localcdn',
        },
    });
});

app.all('/robot/fetch', (req, res) => {
    let robotid = String(
            req.body.robotid ||
            req.params.robotid ||
            '').trim();

    let robot = robotman.load(robotid);

    // never expose key in insecure way
    if (robot.key) delete robot["key"];

    if (req.query.ret == 'json') {
        res.json(robot);
    } else {
        res.send(robot.id || 'none');
    }
});

app.all('/robot/update', (req, res) => {
    if (req.method == "POST") {
        let results = robotman.save(req.body.robot)

        if (req.query.ret == 'json') {
            // return as json
            res.json({
                "now": new Date(),
                "action": results,
            });

        } else {
            // return as html
            if (results[0]) {
                msg = ["green", results[1]];
            } else {
                msg = ["red", results[1]];
            }

            // really simple message "flashing" (set)
            res.cookie(
                'msgResults', msg,
                { maxAge: 60000, httpOnly: true });

            // prevent resubmission prompt on refresh
            res.redirect(req.path);
        }

        return;
    }

    // really simple message "flashing" (get)
    let msgResults = req.cookies.msgResults;
    res.clearCookie('msgResults');

    res.render("robot.html", {
        'msgResults': msgResults,
        'config': {
            'cdn': 'localcdn',
            'regex_name': utils.allowed_patterns.name,
        },
    });
});

// route: arena

app.get("/arena/:arenaid", (req, res) => {
    let arenaid = utils.sanitise_string(req.params.arenaid); // slugified

    // dev: sticky load-balancing for websockets (ws)
    //  nginx, ws has limited "sticky sessions" feature
    //  inject a balancerid key during client socketio init
    //  nginx map reroutes balancerid to upstream
    let balancerid = utils.balancerid(arenaid);

    // generate a robot id and robot key based on the arenaid
    let robotid = 'bot' + arenaid;

    res.render("arena.html", {
        'balancerid': balancerid,
        'arenaid': arenaid,
        'arenaonly': !!req.query.arenaonly,
        'robotid': robotid,
        'robotkey': robotman.calculate_robot_key(robotid),
        'config': {
            'cdn': 'localcdn',
        },
    });
});

app.get("/check/simlog/:arenaid", (req, res) => {
    let arenaid = req.params.arenaid;

    let response = {
        "available": false
    };

    if (check(arenaid)) {
        response["available"] = true;
    }

    res.json(response);
});

// route: main/front page

app.get("/", (req, res) => {
    res.render("index.html", {
        'config': {
            'cdn': 'localcdn',
            'regex_arenaid': utils.allowed_patterns.default
        },
    });
});

// routes: api

app.get("/api", async (req, res) => {
    res.render("api.html", {
        'config': {
            'cdn': 'localcdn',
        },
    });
});

app.all(["/api/status"], (req, res) => {
    res.json(socketman.status(
        (req.body.api || req.query.api) == APIKEY));
});

app.all("/api/close/:arenaid", async (req, res) => {
    if ((req.body.api || req.query.api) != APIKEY) {
        res.status(403);
        res.send();
        return;
    }

    let arenaid = req.params.arenaid,
        response = {
            "now": new Date(),
            "action": false,
        };

    if (process.env.MULTINODE) {
        response["action"] = await socketman.asyncDeregisterArena(arenaid);
    } else {
        response["action"] = socketman.deregisterArena(arenaid);
    }

    res.json(response);
});

app.all("/api/robot/generate", async (req, res) => {
    if ((req.body.api || req.query.api) != APIKEY) {
        res.status(403);
        res.send();
        return;
    }

    res.json({
        "now": new Date(),
        "response": robotman.randomtoken(),
    });
});

app.get(["/api/noop", "/api/noop", "/api/noop/:reflect"], (req, res) => {
    // noop: useful for testing multiprocess deploys

    console.log("noop requested")

    res.json({
        "now": new Date(),
        "response": req.params.reflect || false
    });
});

app.get('/competition/:compid', (req, res) => {
    res.render('competition.html', {
        config: {
            cdn: 'localcdn',
        },
        compid: req.params.compid || false,
    });
})

app.all('/api/competition/:compid', (req, res) => {
    res.json(compman.status(req.params.compid || false));
});

app.all('/api/competition/start/:compid', (req, res) => {
    if ((req.body.api || req.query.api) != APIKEY) {
        res.status(403);
        res.send();
        return;
    }

    console.log(req.body);

    let action = compman.competitionVersus(
        req.params.compid,
        req.body.robotList,
        {
            groupMax: req.body.groupMax
        },{
            maxRounds: req.body.maxRounds,
            autoRestart: req.body.autoRestart,
            limitRound: req.body.limitRound,
        }
    );

    res.json({
        "now": new Date(),
        "action": action,
    });
});

app.all('/test/competition', (req, res) => {
    if ((req.body.api || req.query.api) != APIKEY) {
        res.status(403);
        res.send();
        return;
    }

    let response = compman.competitionVersus(
        'test',
        //'default,\nfastbot,slowbot spinbot spinbot default,\nfastbot,slowbot ' +
        //'spinbot spinbot default,\nfastbot,slowbot spinbot spinbot fastbot' +
        'spinbot spinbot default,\nfastbot,slowbot spinbot spinbot fastbot',
        {
            groupMax: 8
        },{
            maxRounds: 1,
            autoRestart: 3,
        });

    res.json(response);
});
