const utils = require('./utils.js');

const arenaman = require('./arenaman.js');
const compman = require('./compman.js');

let io = false;

let ViewerSockets = {};
let RoboSockets = {};

/* arena, viewers and robot managers */

const registerArena = (data) => {
    let arenaid = data.arenaid,
        Arena = arenaman.get(arenaid);

    let reserved = arenaid.length > 30;

    if (Arena) {
        if (!Arena.settings.sim.locked && !reserved && data.extras.forceNew) {
            // respect locked and reserved arenas and prevent takeovers
            deregisterArena(arenaid);
            Arena = false;
        } else {
            console.log("[sockman] existing arena", arenaid);
        }
    }

    // standard arenas must have ids of 30 chars or less
    if (!Arena && !reserved) {
        let arenaConfig = Object.assign({
            'arenaid': arenaid,
            'clientConnection': () => clientConnector(arenaid),
        }, data.extras || {});

        Arena = arenaman.Arena.multipurpose(arenaConfig)

        if (!Arena.start) {
            // XXX: handle odd error where Arena "goes away"
            // force terminate arena and any sockets
            deregisterArena(arenaid, true)
            return false;
        }

        Arena.start();

        console.log("[sockman] new arena", arenaid);
    }

    if (Arena){
        // prepare to track sockets
        if (!(arenaid in RoboSockets)) RoboSockets[arenaid] = {};
        if (!(arenaid in ViewerSockets)) ViewerSockets[arenaid] = {};
    }

    send_to_parent(instance_status());

    return Arena;
}

const deregisterArena = (arenaid, force) => {
    if (!arenaman.get(arenaid) && !force) return false;

    arenaman.remove(arenaid);

    var count_viewers = terminateSockets(ViewerSockets, arenaid, "keys");
    var count_robots = terminateSockets(RoboSockets, arenaid, "values");

    console.log("[sockman] terminated viewers and robots",
        count_viewers, count_robots)

    if (process.env.MULTINODE) {
        send_to_parent(instance_status());
        send_to_parent({
            "type": "broadcast.notify",
            "payload": {
                action: "socketman.deregisterArena",
                arenaid: arenaid
            }
        });
    }

    return true;
}

const registerRobot = (socket) => {
    let arenaid = socket.handshake.query.arenaid;

    // check: Arena must exist
    var Arena = arenaman.get(arenaid);
    if (!Arena) return false;

    if (!(arenaid in RoboSockets)) RoboSockets[arenaid] = {};

    // auto-assign a 1-indexed debug key in the first available slot
    // limited by Arena.settings.sim.numOfRobots

    let maxCount = Arena.settings.sim.numOfRobots + 1;
    let assignedKey = false;
    for(var i=1; i < maxCount ; i++){
        let robokey = "Debug " + i;
        if (!(robokey in RoboSockets[arenaid])) {
            assignedKey = robokey;
            break;
        }
    }

    if (assignedKey){
        RoboSockets[arenaid][assignedKey] = socket.id;
        console.log("[sockman] robot assigned", arenaid, assignedKey);
        return assignedKey;
    } else {
        console.log("[sockman] no robot slots", arenaid);
        return false;
    }
}

const registerViewer = (socket, arenaid) => {
    // check: Arena must exist
    var Arena = arenaman.get(arenaid);
    if (!Arena) return false;

    if (!(arenaid in ViewerSockets)) ViewerSockets[arenaid] = {};
    ViewerSockets[arenaid][socket.id] = new Date();
}

/* socket connection handlers */

const clientConnector = (arenaid) => {
    return (context_state) => {
        if (!(arenaid in RoboSockets)) {
            throw {
                code: "missing",
                message: "arena id not found"
            }
        }

        if (!(context_state.name in RoboSockets[arenaid])) {
            throw {
                code: "missing",
                message: "robot id not found"
            }
        }

        var socketid = RoboSockets[arenaid][context_state.name];

        var socket = io.sockets.sockets.get(socketid);

        if (!socket) {
            throw {
                code: "missing",
                message: "robot socket invalid"
            }
        }

        return socket;
    }
}

const terminateSockets = (Collection, arenaid, By) => {
    let count = 0;
    if (arenaid in Collection){
        var keys = Object[By](Collection[arenaid]);
        count = keys.length;
        keys.forEach(socketid => {
            var socket = io.sockets.sockets.get(socketid);

            if (socket) socket.disconnect();

            delete Collection[arenaid][socketid];
        });
        delete Collection[arenaid];
    }
    return count;
}

const onRobotConnection = socket => {
    // dev: socket.id changes on each (re)connection
    console.log("[sockman] robot connected", socket.id);

    // construct key from handshake
    // socket.handshake.query.iamarobot
    let robokey = registerRobot(socket);

    socket.emit("robolink");

    socket.on("disconnect", (reason) => {
        let arenaid = socket.handshake.query.arenaid;

        console.log(
            "robot disconnected", arenaid, robokey,
            "reason:", reason);

        if (arenaid in RoboSockets)
            delete RoboSockets[arenaid][robokey];
    });
}

const onViewerConnection = socket => {
    console.log("[sockman] viewer connected", socket.id);

    let arenaid = false,
        Arena = false;

    // from: arena.init.js::connect
    socket.on("register", (data, cb) => {
        Arena = registerArena(data);

        arenaid = data.arenaid;

        socket.join(arenaid);
        registerViewer(socket, arenaid);

        cb({state: Arena.state});
    });

    socket.on("debug all spread", coordinates => {
        Arena.debug_all_spread(coordinates);
    });

    socket.on("debug all rotate left", () => {
        Arena.debug_all_rotate_left();
    });

    socket.on("debug all rotate right", () => {
        Arena.debug_all_rotate_right();
    });

    socket.on("debug all forward", () => {
        Arena.debug_all_move_forward();
    });

    socket.on("debug sim restart", () => {
        Arena.start();
    });

    socket.on("debug sim toggle pause", () => {
        Arena.pause();
    });

    socket.on("debug sim unpause", () => {
        Arena.pause(false);
    });

    socket.on("debug all face center", () => {
        Arena.debug_all_face_center();
    });

    socket.on("debug all fire", () => {
        Arena.debug_all_fire();
    });

    socket.on("disconnect", async (reason) => {
        console.log("viewer disconnected, reason:", reason);

        if (arenaid in ViewerSockets)
            delete ViewerSockets[arenaid][socket.id];
    });
}

const init = socketio => {
    io = socketio;

    arenaman.set_socket_io(io);

    io.on("connection", socket => {
        if (socket.handshake.query.arenaid) {
            // verify we are on the correct balancer
            // dev: reduces issues from load balancer config changes

            let server_balancerid = utils.balancerid(socket.handshake.query.arenaid),
                client_balancerid = socket.handshake.query.balancerid;

            if (server_balancerid != client_balancerid) {
                socket.emit('balancerid.mismatch', {
                    'server': server_balancerid,
                    'client': client_balancerid,
                });
                console.log(
                    '[sockman] mismatched balancerid',
                    server_balancerid, client_balancerid);
            }
        }

        if ("iamarobot" in socket.handshake.query) {
            // from: RobotControlPanel.js::initSocket
            onRobotConnection(socket);
        } else {
            // arena
            onViewerConnection(socket);
        }
    });

    console.log("[sockman] init");
}

const metadata_arenas = () => {
    let collection = [];

    for (const [arenaid, Arena] of Object.entries(arenaman.Arenas)) {
        let metadata = {
            "AID": arenaid,
            "IID": Arena.settings.sim.IID,
            "arenaAge": ((new Date()).getTime() - Arena.settings.sim.created.getTime()) / 1000,
            "created": Arena.settings.sim.created,
            "balancerid": utils.balancerid(arenaid),
        };

        let viewer = {
            "first": false,
            "last": false,
            "count": 0};

        if (arenaid in ViewerSockets) {
            for (const [_, datetime]
                    of Object.entries(ViewerSockets[arenaid])) {

                if (viewer.first === false || viewer.first > datetime) viewer.first = datetime;
                if (viewer.last === false || viewer.last < datetime) viewer.last = datetime;
            };
            viewer.count = Object.keys(ViewerSockets[arenaid]).length;
        }

        metadata["activeViewers"] = viewer;

        collection.push(metadata);
    }

    return collection;
}

const instance_status = () => {
    // !!! returns privileged info

    let status = {
        "now": new Date(),
        "type": "socketman.status",
        "Arenas": metadata_arenas(),
        "RoboSockets": RoboSockets,
        "ViewerSockets": ViewerSockets
    };

    return status;
}

const scrub_status = (status) => {
    let deref = JSON.parse(JSON.stringify(status));

    deref.Arenas.forEach(arena => arena.AID="XXXXXX");
    delete deref.RoboSockets;
    delete deref.ViewerSockets;

    return deref;
}

/* IPC, state-sharing */

// client-client state notifications
// "broadcast.notify" repeats to all clients (inc. sender)
// "notes" expires in 60 seconds
let identifier = "unset",
    GlobalState = false,
    Notes = [];

const set_identifier = (str) => {
    // each instance has a unique arbitrary identifier
    identifier = String(str);
    console.log("[IPC] identifier =", identifier);
}

const check_notes = (func) => {
    for (var i=0; i<Notes.length; i++) {
        var Note = Notes[i];
        if (Note.acknowledged) continue;
        if (func(Note) === false) break;
    }
}

const omni_status = (privileged) => {
    // expose single/multi-aware public status method
    // dev: useful for transparent API that works with both
    //  ensure both return similar data structs

    let ret = false;

    if (GlobalState) {
        ret = GlobalState
    } else {
        ret = instance_status()
    }

    if (privileged) {
        ret['elevated'] = true;
    } else {
        ret = scrub_status(ret);
        ret['elevated'] = false;
    }

    // attach data source
    ret['identifier'] = identifier;

    return ret;
}

const send_to_parent = (data) => {
    if (process.env.MULTINODE) {
        // attach data source
        if (typeof data == 'object') data['identifier'] = identifier;
        process.send(data)
    }
}

const asyncDeregisterArena_check = (arenaid, wait) => {
    return new Promise(resolve => {
      setTimeout(() => {
        let closed = false;
        check_notes((Note) => {
            if (!Note) return;
            if (!Note.payload) return;
            if (Note.payload.action != "socketman.deregisterArena") return;
            if (Note.payload.arenaid == arenaid) {
                Note.acknowledged = (new Date()).getTime();
                closed = true;
                console.log(
                    '[IPC] note ack,',
                    'arena closed:', arenaid);
                return;
            }
        });
        resolve(closed);
      }, wait);
    });
}

const get_port_by_balancer = (balancerid) => {
    if (GlobalState && GlobalState.BalancerPorts) {
        return GlobalState.BalancerPorts[balancerid];
    }
    return false;
}

const asyncDeregisterArena = async (arenaid) => {
    console.log("[IPC] requesting close:", arenaid);
    process.send({
        type: "socketman.deregisterArena",
        arenaid: arenaid
    });
    // wait 2 seconds for a note that arena has closed
    return await asyncDeregisterArena_check(arenaid, 2000);
}

if (process.env.MULTINODE) {
    // forked from multiserver.js
    console.log("[IPC] many instances");

    process.on('message', function(data) {
        if (data.type == "multiserver.status") {
            // receive and store broadcasted state data from parent
            GlobalState = data;
            GlobalState["now"] = new Date();
            GlobalState["type"] = "socketman.status.all";
            GlobalState["identifier"] = identifier;
            // workaround to get balancer ports into compman
            compman.set_balancer_ports(GlobalState.BalancerPorts);
        } else if (data.type == 'socketman.deregisterArena') {
            deregisterArena(data.arenaid);
        } else if (data.type == 'broadcast.notify') {
            if (!('expire' in data)) {
                // default: expire in 60 seconds
                data["expire"] = (new Date()).getTime() + 60000;
            }
            data["acknowledged"] = false;
            Notes.push(data);
        } else {
            console.log("[IPC] unhandled", data);
        }
    });

    setInterval(() => {
        let now = (new Date()).getTime();

        // periodically send full status to parent
        send_to_parent(instance_status());

        // process notes
        check_notes((Note) => {
            if (!Note) return;
            if (!Note.payload) return;
            if (Note.payload.action != "compman.queueArenas") return;
            Note.acknowledged = (new Date()).getTime();

            compman.queueArenas(Note.payload.Matches);
        });

        // remove expired notes
        var temp = [];
        for (var i=0; i<Notes.length; i++) {
            var Note = Notes[i];
            if (now < Note.expire) {
                temp.push(Note);
            } else {
                // dev: noisy when ipc is active
                // console.log('[IPC] IPC note expired', Note);
            }
        }
        Notes = temp;
    }, 1000);

} else {
    console.log("[IPC] single instance");
}

module.exports = {
    RoboSockets: RoboSockets,
    ViewerSockets: ViewerSockets,

    init: init,

    registerArena: registerArena,
    deregisterArena: deregisterArena,

    registerRobot: registerRobot,

    onViewerConnection: onViewerConnection,
    onRobotConnection: onRobotConnection,

    metadata_arenas: metadata_arenas,

    // IPC-related

    set_identifier: set_identifier,
    status: omni_status,
    send_to_parent: send_to_parent,
    get_port_by_balancer: get_port_by_balancer,
    asyncDeregisterArena: asyncDeregisterArena,
};
