
(() => {

var balancerid = window.balancerid || ""; // load-balancer

var io_settings = {
    transports: ["websocket"],
    query: {
        'balancerid': balancerid,
        "arenaid": window.location.pathname.split("/arena/").pop(),
    }
};

const socket = io(io_settings);

const canvas = document.querySelector("#arena");
const ctx = canvas.getContext("2d");

const arenaVectors = ArenaVectorBodies(ctx);
const arenaVisuals = createArenaVisuals(ctx);

let robotById = {},
    serverState = {},
    serverResults = false;

const set_arena_dimensions = (width, height) => {
    canvas.width = width;
    canvas.height = height;
    window.dispatchEvent(new Event('resize')); // dispatch for other handlers
    console.log("reset arena dimensions", width, height);
}

const reset_client_state = () => {
    arenaVisuals.reset();
    serverState = {};
    serverResults = false;
    console.log("reset client state");
}

/* frame renderer */

const render_frame = (robots, walls, projectiles, sensors, extras) => {
    let arena = serverState.arena;

    postrender = []; // dev: optional cleanup

    // start of rendering
    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.clearRect(0, 0, arena.width, arena.height);

    // special: add canvas shake during explosions
    if (extras.explodingRobots.length > 0) {
        var _wall = arena.wall,
            shake_x = Math.random() * _wall - (_wall/2),
            shake_y = Math.random() * _wall - (_wall/2);
        ctx.translate(shake_x, shake_y);
    }

    arenaVisuals.background();

    // physics bodies
    arenaVectors.draw(robots, walls, projectiles, sensors);

    // client-side user experience
    arenaVisuals.render();

    // metadata
    arenaVisuals.hud(serverState);
    arenaVisuals.results(serverState);

    if (serverState.realtime) {
        // end of canvas rendering
        postrender.forEach(f => f());
    }
}

/* notifications */

const show_notification = (Settings) => {
    Settings = Object.assign({}, {
        message: 'No description',
        actions: '',
        css: 'bg-red',
    }, Settings);

    Settings.css = Settings.css.split(" ");

    document.getElementById("notification-message").innerHTML = Settings.message;
    document.getElementById("notification-actions").innerHTML = Settings.actions;

    let notificationBody = document.getElementById("notification-body");
        // remove any extra css from previous notification
        if (notificationBody.dataset.extracss)
            notificationBody.classList.remove(
                ...notificationBody.dataset.extracss.split(' '));
        notificationBody.classList.add(...Settings.css);
        // track the added css if any
        notificationBody.setAttribute("data-extracss", Settings.css.join(' '))

    document.getElementById("main").classList.add("with-notification");
}

/* socket events */

socket.on("connect", () => {
    console.log("connected");
    socket.emit(
        "register", {
            arenaid:
                window.location.pathname.split("/arena/").pop(),
            extras:
                Object.fromEntries(new URLSearchParams(
                    window.location.search.substring(1)))
        }, ({state}) => {
            if (!state) {
                show_notification({
                    message:
                        'Arena unavailable. ' +
                        'Try heading <a class="link dim dark-green" href="javascript: history.back()">back</a> or ' +
                        'the <a class="link dim dark-green" href="/">front page</a> to try again',
                    css: 'black bg-yellow',
                });
                return;
            }
            console.log(state);
            set_arena_dimensions(state.arena.width, state.arena.height);
            reset_client_state();
        }
    ); // socketman.js::onViewerConnection

});

socket.on("reset state", () => {
    reset_client_state();
});

socket.on("robosync", () => {
    console.log("invalid robosync");
});

socket.on("update state", ({robots, walls, projectiles, sensors, status}) => {
    // client-server state sync

    if (status.frame <= serverState.frame) {
        console.log("frame lower", status.frame, serverState.frame);
        reset_client_state();
    }

    serverState = status;

    robotById = [];
    robots.forEach(robot => { robotById[robot.id] = robot });

    // extra states from visual animations
    var extras = arenaVisuals.update(serverState.frame, robotById);

    render_frame(robots, walls, projectiles, sensors, extras);

    // show debug info
    (async function() {
        var debug = robots.map(r =>
            '<div class="mv2 mh0 pr2 db">' +
                // robot name, team, health
                '<div class="ph2 pv1 mb1 white bg-purple db flex justify-between">' +
                    '<div>' +
                        '<div class="f7 b mv1">' + r.entityName + '</div>' +
                        '<div><small>' + r.entityTeam + "</small></div>" +
                    '</div>' +
                    '<div class="flex items-center">' +
                        '<div class="healthbar b--silver bg-black">' +
                            '<div class="bar bg-green" ' +
                                'style="width:' + Math.ceil((Math.max(0, r.health)/5)) + 'px">' +
                            '</div>' +
                        '</div>' +
                    '</div>' +
                '</div>' +
                // list of objects seen by scanner
                '<div class="pl2 mv2">' +
                    r.debugScannerALL.map(e =>
                    '<div class="f7 mb2">' +
                        // e.id + " " +
                        e.entityType + " " +
                        (e.entitySubtype ? '(' + e.entitySubtype + ')' : '') + " " +
                        e.range.toFixed(2) +
                    '</div>'
                    ).join("\n") +
                '</div>' + (
                // brain status (if available)
                (r.brainStatus !== true) ?
                    '<div class="f7 red b tc">' +
                        (
                            (typeof r.brainStatus == 'object' &&
                                r.brainStatus[0] &&
                                r.brainStatus[1]) ?
                                    ("(" + r.brainStatus[0] + ") " + r.brainStatus[1]) :
                                    r.brainStatus
                        ) +
                    '</div>' : ''
                ) +
            '</div>');
        document.getElementById("status-robots").innerHTML = debug.join("\n");
    })();

});

socket.on("robot knockout", ({id}) => {
    if (!serverState.realtime) return;
    arenaVisuals.explode_robot(id);
});

socket.on("simulation stop", () => {
    console.log("simulation stop");
    show_notification({
        message:
            'Simulation stopped. ' +
            (window.location.href.includes("forceNew=1") ?
                '<a class="link dim dark-green" href="javascript: location.reload();" ' +
                    'title="refresh">Refresh page</a> to restart' :
                '<a class="link dim dark-green" href="/">Return to front page</a>'),
        css: 'black bg-light-blue',
    });
});

socket.on('balancerid.mismatch', (data) => {
    console.log("balancer mismatch detected", data);
    show_notification({
        message:
            'Load balancer mismatch. ' +
            'Save any <a class="link dim dark-green" href="#main" title="go to top of page">robots</a> and ' +
            '<a class="link dim dark-green" href="javascript: location.reload();" title="refresh">refresh page</a>',
        css: 'black bg-yellow',
    });
});

/* client-side events for debugging */

/*
document.getElementById("arena").addEventListener("mousedown", e => {
    socket.emit("debug all spread", {x: e.offsetX, y: e.offsetY});
});
*/

document.querySelectorAll("button[data-command]").forEach(el =>{
    el.addEventListener("click", e => { socket.emit(el.dataset.command); });
})

})(); // end: ()
