'use strict';

class RobotControlPanel extends React.Component {
    constructor(props) {
        super(props);

        // first non-blank brain preset
        var brainPreset = props.presets.findIndex(p => !!p[1]),
            brainCode = props.presets[brainPreset][1];

        this.state = {
            minimised: props.minimised || false,
            connected: false,
            status: "init",
            context: "",
            x: false,
            y: false,
            angle: false,
            brainPreset: brainPreset,
            brainCode: brainCode,
            brainStatus: "pending",
            msprocess: 0.0,
        };

        this._isMounted = false;

        this.socket = false;
        this.timeout_status = false;

        this.runtimes = [];

        this.compileBrainFunction(brainCode, true);

        if (props.autoconnect) this.initSocket();
    }

    status_change(status, next) {
        this.setState({status: status});

        clearTimeout(this.timeout_status);
        if (next) this.timeout_status = setTimeout(() => {
            this.status_change(next.status || next);
        }, next.ms || 3000);
    }

    initSocket() {
        var balancerid = window.balancerid || ""; // load-balancer

        this.socket = io({
            transports: ["websocket"],
            forceNew: true,
            query: {
                "iamarobot": true,
                "arenaid": window.location.pathname.split("/arena/").pop(),
                "balancerid": balancerid
            }
        });
        // socketman.js::connection(iamarobot)

        this.socket.on("connect", () => {
            this.status_change("connected", "connected, idle");
            this.setState({connected: true});
        });

        this.socket.on("disconnect", () => {
            if (this._isMounted) {
                // prevent memory leak when socket disconnected during unmount
                this.status_change("disconnected", "not connected");
                this.setState({connected: false});
            }
        });

        this.socket.on("cogitate", (context, cb) => {
            var start = (new Date()).getTime();

            this.status_change("thinking", "connected, idle");

            var _x, _y, _angle;
            try {
                // prevent visual errors from invalid context
                _x = context.state.position.x.toFixed(2);
                _y = context.state.position.y.toFixed(2);
                _angle = context.state.angle.toFixed(3);
            } catch(e) {
                _x = false,
                _y = false,
                _angle = false;
            }

            this.setState({
                x: _x, y: _y, angle: _angle,
                context: JSON.stringify(context, undefined, 2)
            });

            // run brain code
            // dev: only set if it compiles properly
            this.brain_function(context);

            delete context["state"]; // save bandwidth

            cb(context);

            this.runtimes.push((new Date()).getTime() - start);
            this.runtimes = this.runtimes.slice(-10);
            this.setState({
                "msprocess":
                    this.runtimes.reduce((a, b) => a + b, 0)
                        / this.runtimes.length
            });
        });
    }

    disposeSocket() {
        clearTimeout(this.timeout_status);
        this.socket.disconnect();
        this.timeout_status = false;
        this.socket = false;
    }

    compileBrainFunction(code, isinit) {
        let compilerResults = window.doBrainCompile(code);

        // only assign brain function if test completes
        if (compilerResults[0]) this.brain_function = compilerResults[0];

        if (isinit) {
            this.state.brainStatus = compilerResults[1];
        } else {
            this.setState({brainStatus: compilerResults[1]});
        }
    }

    handleSelectPreset(e) {
        var idx = e.target.value,
            code = this.props.presets[idx][1];

        this.compileBrainFunction(code);

        this.setState({
            brainPreset: idx,
            brainCode: code
        });
    }

    handleModifiedBrain(e) {
        var code = (e.target.value + ""),
            idx = this.props.presets
                .findIndex(p => p[1] && p[1].trim() == code);

        if (idx <= 0) {
            // preset not found, use untrusted code
            idx = 0;
        } else {
            // preset found, use that code instead
            code = this.props.presets[idx][1];
        }

        this.compileBrainFunction(code);

        this.setState({
            brainPreset: idx,
            brainCode: code
        });
    }

    handleRemove(e) {
        e.preventDefault();
        e.stopPropagation();
        this.props.onRemoveClick(this);
    }

    componentDidMount() {
        this._isMounted = true;
    }

    componentWillUnmount() {
        this._isMounted = false;
        if (this.state.connected) this.disposeSocket();
    }

    toggleMinimised(e) {
        var minimised = !this.state.minimised;

        this.setState({
            minimised: minimised
        })
    }

    toggleConnect() {
        if (this.state.connected) {
            this.disposeSocket();
        } else {
            if (this.socket === false) {
                this.initSocket();
            } else {
                this.socket.connect();
            }
        }
    }

    doCopyContext() {
        navigator.clipboard.writeText(this.state.context);
    }

    doUploadCode() {
        if (confirm(
                'Your code will be uploaded. ' +
                'Any robot already on the server will be overwritten. ' +
                'Select "OK" to proceed.')) {

            let [robotid, robotkey] = this.props.roboupload;

            fetch('/robot/update?ret=json', {
                method: 'POST',
                mode: 'cors',
                cache: 'no-cache',
                headers: {
                    'Content-Type': 'application/json'
                },
                body: JSON.stringify({
                    robot: {
                        'id': robotid,
                        'key': robotkey,
                        'name': robotid,
                        'team': 'arena',
                        'brain': this.state.brainCode,
                    },
                }),
            })
            .then(res => {
                if (!res.ok) throw Error('Server returned ' + res.status)
                return res.json()
            })
            .then(data => {
                console.log('[doUploadCode] data', data);
                if (data.action) {
                    if (data.action[0]) {
                        alert(
                            data.action[1] + '. ' +
                            'Your robot can be accessed via the Robot ID ' +
                            'shown below the arena.');
                    } else {
                        alert(
                            'There was an issue uploading the robot. ' +
                            'Service returned: ' + data.action[1]);
                    }
                } else {
                    alert('The server did not return a valid response.');
                }
            })
            .catch(err => {
                alert('An error occurred. Details: ' + err);
                console.log('[doRetrieveCode] error', err);
            });
        } // confirm
    }

    doRetrieveCode() {
        let [robotid, _] = this.props.roboupload;

        fetch('/robot/fetch?ret=json', {
            method: 'POST',
            mode: 'cors',
            cache: 'no-cache',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({
                robotid: robotid,
            }),
        })
        .then(res => {
            if (!res.ok) throw Error("Server returned " + res.status)
            return res.json()
        })
        .then(data => {
            if (data.brain) {
                console.log('[doRetrieveCode] data', data);
                if (confirm(
                        'Existing robot found on server. ' +
                        'Press "OK" to replace your current code ' +
                        'or "Cancel" and make a backup first.')) {

                    let code = data.brain || '';

                    this.compileBrainFunction(code);
                    this.setState({
                        brainCode: code,
                        brainPreset: 0,
                    });
                } else {
                    alert(
                        'Your current code is unchanged. ' +
                        'Please make a backup before loading a robot.');
                }
            } else {
                alert(
                    'No existing robot was found. ' +
                    'Your current code is unchanged.');
            }
        })
        .catch(err => {
            alert('An error occurred. Details: ' + err);
            console.log('[doRetrieveCode] error', err);
        });
    }

    render() {
        const e = React.createElement;

        var classNames = ["RobotControlPanel"];
        if (this.state.minimised) classNames.push("minimised");
        classNames = classNames.join(" ");

        // title content
        var titleContent = [
            e('span', {
                key: "title",
                className: "RobotControlPanel__Title",
            }, this.props.name)
        ];
        if (this.props.onRemoveClick) titleContent.push(
            e('span', {
                key: "remove",
                className: "RobotControlPanel__Remove",
                onClick: (e) => this.handleRemove(e)
            }, "X"));

        // buttons
        var robotControls = [
            e('button', {
                key: "connection",
                onClick: () => this.toggleConnect()
            }, this.state.connected ? "disconnect" : "connect"),
            e('button', {
                key: "copy-context",
                onClick: () => this.doCopyContext()
            }, "copy context"),
        ];
        if (this.props.roboupload) robotControls.push(...[
            e('button', {
                key: "save-code",
                onClick: () => this.doUploadCode()
            }, "upload"),
            e('button', {
                key: "load-code",
                onClick: () => this.doRetrieveCode()
            }, "retrieve"),
        ]);

        return e(ReactTransitionGroup.CSSTransition, {
                classNames: "Transition__RobotControlPanel__Robot",
                appear: true,
                in: true,
                timeout: 500,
            }, e('div', {className: classNames}, [
                e('div', {key: "robot", className: "RobotControlPanel__Robot"}, [
                    e('div', {
                        key: "name",
                        className: "RobotControlPanel__Robot__Name",
                        onClick: () => {
                            this.toggleMinimised();
                            if (this.props.onTitleClick)
                                this.props.onTitleClick(this);
                        }
                    }, titleContent),
                    e('div', {key: "status"}, [
                        e('span', {key:"status"}, "status"),
                        (this.state.status === false) ? "unknown": this.state.status
                    ]),
                    e('div', {key: "angle"}, [
                        e('span', {key:"angle"}, "angle"),
                        (this.state.angle === false) ? "?" : this.state.angle
                    ]),
                    e('div', {key: "x"}, [
                        e('span', {key:"x"}, "x"),
                        (this.state.x === false) ? "?":  this.state.x
                    ]),
                    e('div', {key: "y"}, [
                        e('span', {key:"y"}, "y"),
                        (this.state.y === false) ? "?":  this.state.y
                    ]),
                ]),
                e('div', {key: "controls", className: "RobotControlPanel__Controls"}, robotControls),
                e(CollapsiblePanel, {key: "context", className: "RobotControlPanel__Inputs"}, [
                    [
                        e('span', {key: "title"},
                            "Raw Inputs"),
                        e('span', {key: "size"},
                            this.state.context.length + " bytes"),
                    ],
                    e('pre', {key: "json"}, this.state.context || "<no context available>"),
                ]),
                e(CollapsiblePanel, {key: "brain", open:true, className: "RobotControlPanel__Brain"}, [
                    [
                        e('span', {key: "title"},
                            "Brain"),
                        e('span', {key: "status"},
                            this.state.brainStatus),
                        e('span', {key: "timeprocess"},
                            Math.ceil(this.state.msprocess) + "ms"),
                    ],
                    e('textarea', {
                        key: "code",
                        onChange: (e) => this.handleModifiedBrain(e),
                        value: this.state.brainCode || undefined
                    }),
                    e('div', {key: "controls", className: "RobotControlPanel__Brain__Controls"}, [
                        e('select', {
                            key: "presets",
                            value: this.state.brainPreset,
                            onChange: (e) => this.handleSelectPreset(e)
                        }, this.props.presets.map((p, idx) => e("option", {
                            key: idx,
                            value: idx
                        }, p[0]))),
                        e('span', {key:"WIP"}, "")
                    ])
                ]),
            ])
        );
    }
} // RobotControlPanel
